"""Tests for plotter utilities."""

from stanza.plotter.backends.utils import PlotSpec, prepare_heatmap_data


def test_prepare_heatmap_data_with_existing_data():
    """Test prepare_heatmap_data with existing data points."""
    data = {"x": [1.0, 2.0], "y": [1.0, 1.0], "value": [5.0, 10.0]}
    existing_data = {"x": [0.0], "y": [0.0], "value": [3.0]}
    spec = PlotSpec(name="test", plot_type="heatmap", x_label="X", y_label="Y")

    result = prepare_heatmap_data(data, existing_data, spec)

    assert "width" in result
    assert "height" in result
    assert len(result["width"]) == 2
    assert spec.value_min == 5.0
    assert spec.value_max == 10.0


def test_prepare_heatmap_data_empty_key():
    """Test prepare_heatmap_data when a key is missing from new data."""
    data = {"y": [1.0, 2.0], "value": [5.0, 10.0]}
    existing_data = {"x": [0.0], "y": [0.0], "value": [3.0]}
    spec = PlotSpec(name="test", plot_type="heatmap", x_label="X", y_label="Y", dy=0.5)

    result = prepare_heatmap_data(data, existing_data, spec)

    assert spec.dx == 0.1
    assert "width" in result
    assert "height" in result


def test_prepare_heatmap_data_with_preset_cell_size():
    """Test prepare_heatmap_data when cell sizes are already set."""
    data = {"x": [1.0, 2.0], "y": [1.0, 1.0], "value": [5.0, 10.0]}
    existing_data = {}
    spec = PlotSpec(
        name="test", plot_type="heatmap", x_label="X", y_label="Y", dx=0.5, dy=0.3
    )

    result = prepare_heatmap_data(data, existing_data, spec)

    assert spec.dx == 0.5
    assert spec.dy == 0.3
    assert result["width"] == [0.5, 0.5]
    assert result["height"] == [0.3, 0.3]


def test_prepare_heatmap_data_no_value_field():
    """Test prepare_heatmap_data when value field is missing."""
    data = {"x": [1.0, 2.0], "y": [1.0, 1.0]}
    existing_data = {}
    spec = PlotSpec(name="test", plot_type="heatmap", x_label="X", y_label="Y")

    prepare_heatmap_data(data, existing_data, spec)

    assert spec.value_min == float("inf")
    assert spec.value_max == float("-inf")



import pytest

from stanza.base.instruments import BaseControlInstrument, BaseInstrument
from stanza.base.mixins import InstrumentChannelMixin
from stanza.base.registry import load_driver_class, validate_driver_protocols
from stanza.models import InstrumentType


def test_load_driver_class_qdac2():
    driver = load_driver_class("qdac2")
    assert driver.__name__ == "QDAC2"
    assert issubclass(driver, BaseInstrument)


def test_load_driver_class_opx():
    driver = load_driver_class("opx")
    assert driver.__name__ == "OPXInstrument"


def test_load_driver_class_invalid():
    with pytest.raises(ImportError):
        load_driver_class("nonexistent")


def test_validate_driver_protocols_qdac2_general():
    driver = load_driver_class("qdac2")
    validate_driver_protocols(driver, InstrumentType.GENERAL)


def test_validate_driver_protocols_qdac2_control():
    driver = load_driver_class("qdac2")
    validate_driver_protocols(driver, InstrumentType.CONTROL)


def test_validate_driver_protocols_qdac2_measurement():
    driver = load_driver_class("qdac2")
    validate_driver_protocols(driver, InstrumentType.MEASUREMENT)


def test_load_driver_class_no_instrument_found():
    with pytest.raises(ImportError, match="No instrument class found"):
        load_driver_class("utils")


def test_load_driver_class_qswitch():
    driver = load_driver_class("qswitch")
    assert driver.__name__ == "QSwitch"
    assert issubclass(driver, BaseControlInstrument)


def test_validate_driver_protocols_qswitch():
    driver = load_driver_class("qswitch")
    validate_driver_protocols(driver, InstrumentType.BREAKOUT_BOX)


def test_validate_driver_protocols_control_invalid():
    class InvalidControlInstrument(InstrumentChannelMixin):
        pass

    with pytest.raises(TypeError, match="must implement ControlInstrument"):
        validate_driver_protocols(InvalidControlInstrument, InstrumentType.CONTROL)


def test_validate_driver_protocols_measurement_invalid():
    class InvalidMeasurementInstrument(InstrumentChannelMixin):
        pass

    with pytest.raises(TypeError, match="must implement MeasurementInstrument"):
        validate_driver_protocols(
            InvalidMeasurementInstrument, InstrumentType.MEASUREMENT
        )


def test_validate_driver_protocols_general_missing_control():
    class InvalidGeneralInstrument(InstrumentChannelMixin):
        def measure(self, channel_name: str) -> float:
            return 0.0

    with pytest.raises(TypeError, match="must implement ControlInstrument"):
        validate_driver_protocols(InvalidGeneralInstrument, InstrumentType.GENERAL)


def test_validate_driver_protocols_general_missing_measurement():
    class InvalidGeneralInstrument(InstrumentChannelMixin):
        def set_voltage(self, channel_name: str, voltage: float) -> None:
            pass

        def get_voltage(self, channel_name: str) -> float:
            return 0.0

        def get_slew_rate(self, channel_name: str) -> float:
            return 1.0

    with pytest.raises(TypeError, match="must implement MeasurementInstrument"):
        validate_driver_protocols(InvalidGeneralInstrument, InstrumentType.GENERAL)


def test_validate_driver_protocols_breakout_box_invalid():
    class InvalidBreakoutBoxInstrument(InstrumentChannelMixin):
        pass

    with pytest.raises(TypeError, match="must implement BreakoutBoxInstrument"):
        validate_driver_protocols(
            InvalidBreakoutBoxInstrument, InstrumentType.BREAKOUT_BOX
        )



"""Shared utilities for plotting backends.

This module centralizes small, reusable helpers used by multiple
plotting backends to keep implementations minimal and consistent.
"""

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, Any

import numpy as np
from bokeh.models import ColumnDataSource, LinearColorMapper
from bokeh.models.annotations import ColorBar
from bokeh.palettes import Viridis256
from bokeh.plotting import figure

if TYPE_CHECKING:
    pass

# Plot dimensions
LINE_PLOT_WIDTH = 800
LINE_PLOT_HEIGHT = 400
HEATMAP_WIDTH = 800
HEATMAP_HEIGHT = 600


@dataclass
class PlotSpec:
    """Plot specification with type-safe fields."""

    name: str
    plot_type: str  # "line" or "heatmap"
    x_label: str
    y_label: str
    z_label: str | None = None
    cell_size: tuple[float, float] | None = None

    # Heatmap-specific fields
    mapper: LinearColorMapper | None = None
    dx: float | None = None
    dy: float | None = None
    value_min: float = float("inf")
    value_max: float = float("-inf")


@dataclass
class PlotState:
    """Complete plot state: source, figure, and spec in one place.

    Single source of truth for a plot's data, visualization, and configuration.
    Ownership is clear: whoever has the PlotState owns the entire plot.
    """

    source: ColumnDataSource
    figure: Any  # bokeh Figure
    spec: PlotSpec


def prepare_heatmap_data(
    data: dict[str, Any],
    existing_data: dict[str, Any],
    spec: PlotSpec,
) -> dict[str, Any]:
    """Calculate rect sizes and update color range for heatmap data.

    Args:
        data: New data to add (x, y, value)
        existing_data: Existing data in the data source
        spec: Plot specification with dx, dy, value_min, value_max

    Returns:
        Updated data with width and height fields added
    """

    def calc_delta(key: str) -> float:
        """Calculate minimum delta from existing + new data."""
        if key not in data or len(data[key]) == 0:
            return 0.1
        existing_vals = list(existing_data.get(key, []))
        all_vals = existing_vals + data[key]
        if len(all_vals) > 1:
            unique = sorted(set(all_vals))
            if len(unique) > 1:
                return float(min(np.diff(unique)))
        return 0.1

    # Calculate cell sizes if not set
    if spec.dx is None:
        spec.dx = calc_delta("x")
    if spec.dy is None:
        spec.dy = calc_delta("y")

    # Add width/height to data
    n = len(data.get("value", []))
    data["width"] = [spec.dx] * n
    data["height"] = [spec.dy] * n

    # Update value range for color mapping
    if "value" in data:
        values = np.array(data["value"])
        spec.value_min = min(spec.value_min, float(values.min()))
        spec.value_max = max(spec.value_max, float(values.max()))

    return data


def make_line_plot(
    name: str,
    x_label: str,
    y_label: str,
) -> PlotState:
    """Create a 1D line plot.

    Returns complete PlotState with source, figure, and spec.
    """
    source = ColumnDataSource(data={"x": [], "y": []})
    fig = figure(title=name, width=LINE_PLOT_WIDTH, height=LINE_PLOT_HEIGHT)
    fig.xaxis.axis_label = x_label
    fig.yaxis.axis_label = y_label
    fig.line("x", "y", source=source, line_width=2, color="navy")

    spec = PlotSpec(
        name=name,
        plot_type="line",
        x_label=x_label,
        y_label=y_label,
    )
    return PlotState(source=source, figure=fig, spec=spec)


def make_heatmap_plot(
    name: str,
    x_label: str,
    y_label: str,
    z_label: str | None,
    cell_size: tuple[float, float] | None,
) -> PlotState:
    """Create a 2D heatmap plot.

    Returns complete PlotState with source, figure, and spec.
    """
    source = ColumnDataSource(
        data={"x": [], "y": [], "value": [], "width": [], "height": []}
    )

    mapper = LinearColorMapper(palette=Viridis256, low=0, high=1)

    fig = figure(title=name, width=HEATMAP_WIDTH, height=HEATMAP_HEIGHT)
    fig.xaxis.axis_label = x_label
    fig.yaxis.axis_label = y_label

    fig.rect(
        x="x",
        y="y",
        width="width",
        height="height",
        source=source,
        fill_color={"field": "value", "transform": mapper},
        line_color=None,
    )

    color_bar = ColorBar(
        color_mapper=mapper,
        width=8,
        location=(0, 0),
        title=z_label or "Value",
    )
    fig.add_layout(color_bar, "right")

    spec = PlotSpec(
        name=name,
        plot_type="heatmap",
        x_label=x_label,
        y_label=y_label,
        z_label=z_label,
        cell_size=cell_size,
        mapper=mapper,
        dx=cell_size[0] if cell_size else None,
        dy=cell_size[1] if cell_size else None,
    )
    return PlotState(source=source, figure=fig, spec=spec)



from __future__ import annotations

import logging
from typing import overload

from stanza.base.channels import ChannelConfig, InstrumentChannel, Parameter
from stanza.base.instruments import BaseControlInstrument
from stanza.models import ControlInstrumentConfig
from stanza.pyvisa import PyVisaDriver

logger = logging.getLogger(__name__)


class QSwitchChannel(InstrumentChannel):
    """QSwitch channel for setting and getting the state of the lines and relays."""

    def __init__(
        self, name: str, channel_id: int, config: ChannelConfig, driver: PyVisaDriver
    ):
        self.name = name
        self.driver = driver
        super().__init__(config)
        self.channel_id = channel_id

    def _query_all_relays(self, command: str) -> list[int]:
        """Query all 10 relays on this channel and return their states as a list of integers."""
        relays = ",".join(f"{self.channel_id}!{i}" for i in range(10))
        response = self.driver.query(f"{command}? (@{relays})")
        return [int(v) for v in response.split(",")]

    def _set_relay(self, command: str, relay: int | str) -> None:
        """Execute a relay command for a single relay (int) or multiple relays (comma-separated string)."""
        if isinstance(relay, int):
            self.driver.write(f"{command} (@{self.channel_id}!{relay})")
        else:
            relays = ",".join(f"(@{self.channel_id}!{r})" for r in relay.split(","))
            self.driver.write(f"{command} (@{relays})")

    def _get_connect(self) -> list[int]:
        """Get the connection state of all relays on this channel."""
        return self._query_all_relays("close")

    def _set_connect(self, relay: int | str) -> None:
        """Connect the specified relay or relays on this channel."""
        self._set_relay("close", relay)

    def _get_disconnect(self) -> list[int]:
        """Get the disconnection state of all relays on this channel."""
        return self._query_all_relays("open")

    def _set_disconnect(self, relay: int | str) -> None:
        """Disconnect the specified relay or relays on this channel."""
        self._set_relay("open", relay)

    def _setup_parameters(self) -> None:
        """Initialize and add connect_relay and disconnect_relay parameters to this channel."""
        self.add_parameter(
            Parameter(
                name="connect_relay",
                value=None,
                unit="bool",
                getter=self._get_connect,
                setter=self._set_connect,
            )
        )

        self.add_parameter(
            Parameter(
                name="disconnect_relay",
                value=None,
                unit="bool",
                getter=self._get_disconnect,
                setter=self._set_disconnect,
            )
        )


class QSwitch(BaseControlInstrument):
    def __init__(
        self,
        instrument_config: ControlInstrumentConfig,
        channel_configs: dict[str, ChannelConfig],
        is_simulation: bool = False,
        sim_file: str | None = None,
    ):
        self.name = instrument_config.name
        self.address = instrument_config.ip_addr or instrument_config.serial_addr
        self.port = instrument_config.port
        self.channel_configs = channel_configs

        if is_simulation:
            visa_addr = "ASRL2::INSTR"
            logger.info("Using simulation mode for QSwitch")
        else:
            visa_addr = f"TCPIP::{self.address}::{self.port}::SOCKET"

        self.driver = PyVisaDriver(visa_addr, sim_file=sim_file)
        super().__init__(instrument_config)
        self._initialize_channels(channel_configs)

    def _initialize_channels(self, channel_configs: dict[str, ChannelConfig]) -> None:
        """Create and add QSwitchChannel instances for all configured breakout channels."""
        for cfg in channel_configs.values():
            if cfg.breakout_channel is not None:
                self.add_channel(
                    cfg.name,
                    QSwitchChannel(cfg.name, cfg.breakout_channel, cfg, self.driver),
                )

    def _channels_suffix(self, channel_names: list[str], tap: int | str) -> str:
        """Format channel names and tap number into a SCPI command suffix for multi-channel operations."""
        channels = ",".join(
            f"{self.channel_configs[ch].breakout_channel}!{tap}" for ch in channel_names
        )
        return f"(@{channels})"

    @overload
    def get_grounded(self, channel_name: list[str]) -> list[bool]: ...

    @overload
    def get_grounded(self, channel_name: str) -> bool: ...

    def get_grounded(self, channel_name: str | list[str]) -> bool | list[bool]:
        """Check if breakout box channel(s) are grounded."""
        return self.get_connected(channel_name, 0)

    @overload
    def set_grounded(self, channel_name: list[str]) -> None: ...

    @overload
    def set_grounded(self, channel_name: str) -> None: ...

    def set_grounded(self, channel_name: str | list[str]) -> None:
        """Set breakout box channel(s) to grounded."""
        self.set_connected(channel_name, 0)

    @overload
    def get_ungrounded(self, channel_name: list[str]) -> list[bool]: ...

    @overload
    def get_ungrounded(self, channel_name: str) -> bool: ...

    def get_ungrounded(self, channel_name: str | list[str]) -> bool | list[bool]:
        """Check if breakout box channel(s) are ungrounded."""
        return self.get_disconnected(channel_name, 0)

    @overload
    def set_ungrounded(self, channel_name: list[str]) -> None: ...

    @overload
    def set_ungrounded(self, channel_name: str) -> None: ...

    def set_ungrounded(self, channel_name: str | list[str]) -> None:
        """Set breakout box channel(s) to ungrounded."""
        self.set_disconnected(channel_name, 0)

    @overload
    def get_connected(
        self, channel_name: list[str], line_number: int
    ) -> list[bool]: ...

    @overload
    def get_connected(self, channel_name: str, line_number: int) -> bool: ...

    def get_connected(
        self, channel_name: str | list[str], line_number: int
    ) -> bool | list[bool]:
        """Get if breakout box channel(s) are connected to the line number."""
        if isinstance(channel_name, str):
            return bool(
                self.get_channel(channel_name).get_parameter_value("connect_relay")[
                    line_number
                ]
            )

        response = self.driver.query(
            f"close? {self._channels_suffix(channel_name, line_number)}"
        )
        return [bool(int(v)) for v in response.split(",")]

    @overload
    def set_connected(self, channel_name: list[str], line_number: int) -> None: ...

    @overload
    def set_connected(self, channel_name: str, line_number: int) -> None: ...

    def set_connected(self, channel_name: str | list[str], line_number: int) -> None:
        """Set breakout box channel(s) to connected to the line number."""
        if isinstance(channel_name, str):
            self.get_channel(channel_name).set_parameter("connect_relay", line_number)
        else:
            self.driver.write(
                f"close {self._channels_suffix(channel_name, line_number)}"
            )

    @overload
    def get_disconnected(
        self, channel_name: list[str], line_number: int
    ) -> list[bool]: ...

    @overload
    def get_disconnected(self, channel_name: str, line_number: int) -> bool: ...

    def get_disconnected(
        self, channel_name: str | list[str], line_number: int
    ) -> bool | list[bool]:
        """Get if breakout box channel(s) are disconnected from the line number."""
        if isinstance(channel_name, str):
            return bool(
                self.get_channel(channel_name).get_parameter_value("disconnect_relay")[
                    line_number
                ]
            )

        response = self.driver.query(
            f"open? {self._channels_suffix(channel_name, line_number)}"
        )
        return [bool(int(v)) for v in response.split(",")]

    @overload
    def set_disconnected(self, channel_name: list[str], line_number: int) -> None: ...

    @overload
    def set_disconnected(self, channel_name: str, line_number: int) -> None: ...

    def set_disconnected(self, channel_name: str | list[str], line_number: int) -> None:
        """Set breakout box channel(s) to disconnected from the line number."""
        if isinstance(channel_name, str):
            self.get_channel(channel_name).set_parameter(
                "disconnect_relay", line_number
            )
        else:
            self.driver.write(
                f"open {self._channels_suffix(channel_name, line_number)}"
            )

    def close(self) -> None:
        """Close the QSwitch driver."""
        self.driver.close()



from typing import Protocol, overload, runtime_checkable


@runtime_checkable
class BreakoutBoxInstrument(Protocol):
    """Protocol for a digital breakout box."""

    @overload
    def get_grounded(self, channel_name: list[str]) -> list[bool]:
        """Check if breakout box channels are grounded."""
        ...

    @overload
    def get_grounded(self, channel_name: str) -> bool:
        """Check if breakout box channels are grounded."""
        ...

    def get_grounded(self, channel_name: str | list[str]) -> bool | list[bool]:
        """Check if breakout box channel(s) are grounded."""
        ...

    @overload
    def set_grounded(self, channel_name: list[str]) -> None:
        """Set breakout box channels to grounded."""
        ...

    @overload
    def set_grounded(self, channel_name: str) -> None:
        """Set breakout box channel to grounded."""
        ...

    def set_grounded(self, channel_name: str | list[str]) -> None:
        """Set breakout box channel(s) to grounded."""
        ...

    @overload
    def get_ungrounded(self, channel_name: list[str]) -> list[bool]:
        """Check if breakout box channels are ungrounded."""
        ...

    @overload
    def get_ungrounded(self, channel_name: str) -> bool:
        """Check if breakout box channel is ungrounded."""
        ...

    def get_ungrounded(self, channel_name: str | list[str]) -> bool | list[bool]:
        """Check if breakout box channel(s) are ungrounded."""
        ...

    @overload
    def set_ungrounded(self, channel_name: list[str]) -> None:
        """Set breakout box channels to ungrounded."""
        ...

    @overload
    def set_ungrounded(self, channel_name: str) -> None:
        """Set breakout box channel to ungrounded."""
        ...

    def set_ungrounded(self, channel_name: str | list[str]) -> None:
        """Set breakout box channel(s) to ungrounded."""
        ...

    @overload
    def get_connected(self, channel_name: list[str], line_number: int) -> list[bool]:
        """Check if breakout box channels are connected to the line number."""
        ...

    @overload
    def get_connected(self, channel_name: str, line_number: int) -> bool:
        """Check if breakout box channel is connected to the line number."""
        ...

    def get_connected(
        self, channel_name: str | list[str], line_number: int
    ) -> bool | list[bool]:
        """Get if breakout box channel(s) are connected to the line number."""
        ...

    @overload
    def set_connected(self, channel_name: list[str], line_number: int) -> None:
        """Set breakout box channels to connected to the line number."""
        ...

    @overload
    def set_connected(self, channel_name: str, line_number: int) -> None:
        """Set breakout box channel to connected to the line number."""
        ...

    def set_connected(self, channel_name: str | list[str], line_number: int) -> None:
        """Set breakout box channel(s) to connected to the line number."""
        ...

    @overload
    def get_disconnected(self, channel_name: list[str], line_number: int) -> list[bool]:
        """Check if breakout box channels are disconnected from the line number."""
        ...

    @overload
    def get_disconnected(self, channel_name: str, line_number: int) -> bool:
        """Check if breakout box channel is disconnected from the line number."""
        ...

    def get_disconnected(
        self, channel_name: str | list[str], line_number: int
    ) -> bool | list[bool]:
        """Get if breakout box channel(s) are disconnected from the line number."""
        ...

    @overload
    def set_disconnected(self, channel_name: list[str], line_number: int) -> None:
        """Set breakout box channels to disconnected from the line number."""
        ...

    @overload
    def set_disconnected(self, channel_name: str, line_number: int) -> None:
        """Set breakout box channel to disconnected from the line number."""
        ...

    def set_disconnected(self, channel_name: str | list[str], line_number: int) -> None:
        """Set breakout box channel(s) to disconnected from the line number."""
        ...


@runtime_checkable
class ControlInstrument(Protocol):
    """Protocol for control instruments."""

    def set_voltage(self, channel_name: str, voltage: float) -> None:
        """Set the voltage on a specific channel."""
        ...

    def get_voltage(self, channel_name: str) -> float:
        """Get the voltage on a specific channel."""
        ...

    def get_slew_rate(self, channel_name: str) -> float:
        """Get the slew rate on a specific channel."""
        ...


@runtime_checkable
class MeasurementInstrument(Protocol):
    """Protocol for measurement instruments."""

    @overload
    def measure(self, channel_name: str) -> float:
        """Measure the current on a specific channel."""
        ...

    @overload
    def measure(self, channel_name: list[str]) -> list[float]:
        """Optional overload for measuring the current on multiple channels."""
        ...

    def measure(self, channel_name: str | list[str]) -> float | list[float]:
        """Measure the current on a specific channel(s)."""
        ...


class NamedResource(Protocol):
    """Protocol for resources that have a name attribute."""

    name: str



"""Tests for 2D sweep functionality."""

import tempfile

import numpy as np
import pytest

from stanza.logger.data_logger import DataLogger
from stanza.logger.datatypes import SweepData


def test_sweep_context_2d_dimension_detection():
    """Test that dimension is detected from first append."""
    tmpdir = tempfile.mkdtemp()
    logger = DataLogger(routine_name="test_2d", base_dir=tmpdir)
    session = logger.create_session()

    # 1D sweep
    with session.sweep("1d", "X", "Y") as s:
        s.append([1.0], [2.0])
        assert s._dim == 1

    # 2D sweep
    with session.sweep("2d", ["X1", "X2"], "Y") as s:
        s.append([1.0, 2.0], [3.0])
        assert s._dim == 2

    session.close()


def test_sweep_context_2d_dimension_consistency():
    """Test that dimension is enforced after first append."""
    tmpdir = tempfile.mkdtemp()
    logger = DataLogger(routine_name="test_dim_consistency", base_dir=tmpdir)
    session = logger.create_session()

    # Start with 2D
    with pytest.raises(ValueError, match="2D requires len"):
        with session.sweep("inconsistent", ["X1", "X2"], "Y") as s:
            s.append([1.0, 2.0], [3.0])  # Sets dim=2
            s.append([1.0], [2.0])  # Mismatch!

    session.close()


def test_sweep_context_2d_data_accumulation():
    """Test that 2D data accumulates correctly."""
    tmpdir = tempfile.mkdtemp()
    logger = DataLogger(routine_name="test_accumulation", base_dir=tmpdir)
    session = logger.create_session()

    with session.sweep("2d_accum", ["V1", "V2"], "I") as s:
        for v1 in [-1.0, 0.0, 1.0]:
            for v2 in [-0.5, 0.5]:
                i = v1**2 + v2**2
                s.append([v1, v2], [i])

    # Verify data was accumulated
    assert len(s._x_data) == 6
    assert len(s._y_data) == 6

    # Verify first and last points
    assert s._x_data[0] == [-1.0, -0.5]
    assert s._x_data[-1] == [1.0, 0.5]

    session.close()


def test_sweep_context_2d_validation():
    """Test validation of 2D sweep data."""
    tmpdir = tempfile.mkdtemp()
    logger = DataLogger(routine_name="test_validation", base_dir=tmpdir)
    session = logger.create_session()

    # Test NaN rejection
    with pytest.raises(ValueError, match="NaN or inf"):
        with session.sweep("nan_test", ["X1", "X2"], "Y") as s:
            s.append([1.0, np.nan], [2.0])

    # Test empty data rejection
    with pytest.raises(ValueError, match="cannot be empty"):
        with session.sweep("empty_test", ["X1", "X2"], "Y") as s:
            s.append([], [])

    # Test y must be scalar
    with pytest.raises(ValueError, match="y_data must be length 1"):
        with session.sweep("y_scalar_test", ["X1", "X2"], "Y") as s:
            s.append([1.0, 2.0], [3.0, 4.0])

    # Test 3D not supported
    with pytest.raises(ValueError, match="Only 1D and 2D supported"):
        with session.sweep("3d_test", ["X1", "X2", "X3"], "Y") as s:
            s.append([1.0, 2.0, 3.0], [4.0])

    session.close()


def test_sweep_data_2d_validation():
    """Test SweepData validation for 2D data."""
    # Valid 2D sweep
    sweep = SweepData(
        name="test",
        x_data=np.array([[1.0, 2.0], [3.0, 4.0]]),
        y_data=np.array([5.0, 6.0]),
        x_label=["X1", "X2"],
        y_label="Y",
        metadata={},
        timestamp=0.0,
        session_id="test",
    )
    assert sweep.x_data.shape == (2, 2)

    # Invalid: x_label length mismatch
    with pytest.raises(ValueError, match="x_label length"):
        SweepData(
            name="test",
            x_data=np.array([[1.0, 2.0], [3.0, 4.0]]),
            y_data=np.array([5.0, 6.0]),
            x_label=["X1"],  # Should be 2 labels!
            y_label="Y",
            metadata={},
            timestamp=0.0,
            session_id="test",
        )

    # Invalid: y_data not 1D
    with pytest.raises(ValueError, match="y_data must be 1D"):
        SweepData(
            name="test",
            x_data=np.array([[1.0, 2.0]]),
            y_data=np.array([[5.0]]),  # 2D!
            x_label=["X1", "X2"],
            y_label="Y",
            metadata={},
            timestamp=0.0,
            session_id="test",
        )

    # Invalid: length mismatch
    with pytest.raises(ValueError, match="Length mismatch"):
        SweepData(
            name="test",
            x_data=np.array([[1.0, 2.0], [3.0, 4.0]]),
            y_data=np.array([5.0]),  # Wrong length!
            x_label=["X1", "X2"],
            y_label="Y",
            metadata={},
            timestamp=0.0,
            session_id="test",
        )


def test_sweep_context_2d_with_live_plotting():
    """Test 2D sweep with Bokeh live plotting."""
    from stanza.plotter import enable_live_plotting

    tmpdir = tempfile.mkdtemp()
    logger = DataLogger(routine_name="test_2d_live", base_dir=tmpdir)

    # Enable inline backend (mock - won't actually display)
    try:
        _backend = enable_live_plotting(logger, backend="inline")
    except ImportError:
        pytest.skip("jupyter_bokeh not installed")

    session = logger.create_session()

    # Stream 2D data
    with session.sweep("heatmap", ["Gate 1 (V)", "Gate 2 (V)"], "Current (A)") as s:
        for v1 in np.linspace(-1, 1, 5):
            for v2 in np.linspace(-1, 1, 5):
                current = v1**2 + v2**2
                s.append([v1, v2], [current])

    # Verify sweep completed
    assert len(s._x_data) == 25
    assert len(s._y_data) == 25

    session.close()


def test_sweep_context_2d_metadata():
    """Test that dimension metadata is stored correctly."""
    tmpdir = tempfile.mkdtemp()
    logger = DataLogger(routine_name="test_metadata", base_dir=tmpdir)
    session = logger.create_session()

    with session.sweep("2d_meta", ["X1", "X2"], "Y") as s:
        s.append([1.0, 2.0], [3.0])
        assert s.metadata["_dim"] == 2

    session.close()


def test_sweep_context_2d_cell_size_override():
    """Test cell_size metadata for explicit rect sizing."""
    tmpdir = tempfile.mkdtemp()
    logger = DataLogger(routine_name="test_cell_size", base_dir=tmpdir)
    session = logger.create_session()

    # Provide explicit cell size
    with session.sweep(
        "2d_sized",
        ["X1", "X2"],
        "Y",
        metadata={"cell_size": (0.01, 0.02)},
    ) as s:
        s.append([1.0, 2.0], [3.0])
        assert s.metadata["cell_size"] == (0.01, 0.02)

    session.close()


def test_sweep_context_2d_empty_no_write():
    """Test that empty 2D sweep doesn't write."""
    tmpdir = tempfile.mkdtemp()
    logger = DataLogger(routine_name="test_empty", base_dir=tmpdir)
    session = logger.create_session()

    with session.sweep("empty_2d", ["X1", "X2"], "Y") as _s:
        pass  # No appends

    # Should not raise, just skip
    session.close()


def test_sweep_context_2d_cancel():
    """Test canceling a 2D sweep."""
    tmpdir = tempfile.mkdtemp()
    logger = DataLogger(routine_name="test_cancel", base_dir=tmpdir)
    session = logger.create_session()

    with session.sweep("cancel_2d", ["X1", "X2"], "Y") as s:
        s.append([1.0, 2.0], [3.0])
        s.cancel()

    # Data should be cleared
    assert len(s._x_data) == 0
    assert len(s._y_data) == 0
    assert not s._active

    session.close()


def test_sweep_context_2d_exception_no_write():
    """Test that exception in context prevents write."""
    tmpdir = tempfile.mkdtemp()
    logger = DataLogger(routine_name="test_exception", base_dir=tmpdir)
    session = logger.create_session()

    try:
        with session.sweep("exception_2d", ["X1", "X2"], "Y") as s:
            s.append([1.0, 2.0], [3.0])
            raise RuntimeError("Test error")
    except RuntimeError:
        pass

    # Sweep should have been canceled
    assert not s._active

    session.close()


def test_bokeh_writer_dimension_detection():
    """Test that BokehLiveWriter correctly detects 1D vs 2D."""
    from stanza.logger.writers.bokeh_writer import BokehLiveWriter
    from stanza.plotter.backends.server import ServerBackend

    backend = ServerBackend(port=5007)
    backend.start()

    writer = BokehLiveWriter(backend=backend, max_points=1000)
    writer.initialize_session(None)

    # 1D sweep
    sweep_1d = SweepData(
        name="1d_test",
        x_data=np.array([1.0, 2.0, 3.0]),
        y_data=np.array([4.0, 5.0, 6.0]),
        x_label="X",
        y_label="Y",
        metadata={"_dim": 1},
        timestamp=0.0,
        session_id="test",
    )
    writer.write_sweep(sweep_1d)
    assert writer._plots["1d_test"]["dim"] == 1

    # 2D sweep
    sweep_2d = SweepData(
        name="2d_test",
        x_data=np.array([[1.0, 2.0], [3.0, 4.0]]),
        y_data=np.array([5.0, 6.0]),
        x_label=["X1", "X2"],
        y_label="Y",
        metadata={"_dim": 2},
        timestamp=0.0,
        session_id="test",
    )
    writer.write_sweep(sweep_2d)
    assert writer._plots["2d_test"]["dim"] == 2

    backend.stop()


def test_sweep_context_active_sweep_collision():
    """Test that same-name sweeps cannot be active simultaneously."""
    tmpdir = tempfile.mkdtemp()
    logger = DataLogger(routine_name="test_collision", base_dir=tmpdir)
    session = logger.create_session()

    with session.sweep("test", ["X1", "X2"], "Y") as _s1:
        # Try to start another with same name
        with pytest.raises(Exception, match="already active"):
            with session.sweep("test", ["X1", "X2"], "Y") as _s2:
                pass

    # After first ends, should be able to start another
    with session.sweep("test", ["X1", "X2"], "Y") as s2:
        s2.append([1.0, 2.0], [3.0])

    session.close()


if __name__ == "__main__":
    pytest.main([__file__, "-v"])



from __future__ import annotations

import logging
from pathlib import Path

try:
    import h5py  # type: ignore[import-untyped]
except ImportError:
    h5py = None

import numpy as np

from stanza.exceptions import WriterError
from stanza.logger.datatypes import MeasurementData, SessionMetadata, SweepData
from stanza.logger.writers.base import AbstractDataWriter
from stanza.timing import to_epoch

logger = logging.getLogger(__name__)


class HDF5Writer(AbstractDataWriter):
    """Writer for HDF5 files."""

    def __init__(
        self,
        base_directory: str | Path,
        compression: str | None = None,
        compression_level: int | None = None,
    ):
        if h5py is None:
            raise ImportError(
                "h5py is not installed. Install with: pip install stanza[hdf5]"
            )
        self.base_directory = Path(base_directory)
        if not self.base_directory.exists():
            self.base_directory.mkdir(parents=True, exist_ok=True)

        self.compression = compression
        self.compression_level = compression_level
        self._session_file: Path | None = None
        self._h5_file: h5py.File | None = None
        self._measurement_counters: dict[str, int] = {}

        logger.info(
            "Initialized HDF5 writer with base directory: %s", self.base_directory
        )

    def initialize_session(self, session: SessionMetadata) -> None:
        """Initialize the writer for a new session.

        Args:
            session: Session metadata to write

        Raises:
            WriterError: If session is already initialized or file creation fails
        """
        try:
            self.session_id = session.session_id
            session_filename = f"{self.session_id}.h5"
            self._session_file = self.base_directory / session_filename

            # Open HDF5 file and write session metadata
            self._h5_file = h5py.File(self._session_file, "w")

            metadata_group = self._h5_file.create_group("metadata")
            metadata_group.attrs["routine_name"] = (
                session.routine_name if session.routine_name is not None else ""
            )
            metadata_group.attrs["start_time"] = session.start_time
            metadata_group.attrs["session_id"] = session.session_id
            metadata_group.attrs["device_config"] = str(session.device_config)

            self._h5_file.create_group("measurements")
            self._h5_file.create_group("sweeps")
            self._h5_file.create_group("analysis")

            logger.info("Initialized HDF5 writer for session: %s", self.session_id)
        except Exception as e:
            if self._h5_file is not None:
                self._h5_file.close()
                self._h5_file = None
            self._session_file = None
            raise WriterError(
                f"Failed to initialize HDF5 writer for session: {self.session_id}"
            ) from e

    def finalize_session(self, session: SessionMetadata | None = None) -> None:
        """Finalize the writer for a session.

        Args:
            session: Optional updated session metadata to write

        Raises:
            WriterError: If no active session or finalization fails
        """
        if self._session_file is None:
            raise WriterError("No active session")

        try:
            if session is not None and self._h5_file is not None:
                metadata_group = self._h5_file["metadata"]
                if session.end_time is not None:
                    metadata_group.attrs["end_time"] = session.end_time
                if session.parameters is not None:
                    metadata_group.attrs["parameters"] = str(session.parameters)

            if hasattr(self, "_h5_file") and self._h5_file is not None:
                self._h5_file.close()
                self._h5_file = None

            logger.debug("Session finalized: %s", self._session_file)
            self._session_file = None
            self._measurement_counters.clear()

        except Exception as e:
            raise WriterError(
                f"Failed to finalize HDF5 writer for session: {self.session_id}"
            ) from e

    def write_measurement(self, measurement: MeasurementData) -> None:
        """Write a single measurement data point.

        Args:
            measurement: Measurement data to write

        Raises:
            WriterError: If no active session or write operation fails
        """
        if self._session_file is None or self._h5_file is None:
            raise WriterError("No active session")

        try:
            measurement_name = measurement.name
            if measurement_name in self._measurement_counters:
                counter = self._measurement_counters[measurement_name]
                self._measurement_counters[measurement_name] += 1
                measurement_name = f"{measurement_name}_{counter}"
            else:
                self._measurement_counters[measurement_name] = 1

            is_analysis = measurement.metadata.get("data_type") == "analysis"
            target_group = self._h5_file["analysis" if is_analysis else "measurements"]

            measurement_group = target_group.create_group(measurement_name)

            data_group = measurement_group.create_group("data")
            metadata_group = measurement_group.create_group("metadata")

            for key, value in measurement.data.items():
                if isinstance(value, (np.ndarray, list, tuple)):
                    self._create_dataset(data_group, key, np.array(value))
                else:
                    data_group.attrs[key] = value

            for key, value in measurement.metadata.items():
                try:
                    metadata_group.attrs[key] = value
                except (TypeError, ValueError):
                    metadata_group.attrs[key] = str(value)

            # Store measurement-level metadata
            metadata_group.attrs["timestamp"] = to_epoch(measurement.timestamp)
            metadata_group.attrs["session_id"] = measurement.session_id
            metadata_group.attrs["routine_name"] = (
                measurement.routine_name if measurement.routine_name is not None else ""
            )

            logger.info("Wrote measurement: %s", measurement_name)
        except Exception as e:
            raise WriterError(f"Failed to write measurement: {measurement_name}") from e

    def write_sweep(self, sweep: SweepData) -> None:
        """Write sweep data to HDF5 file.

        Args:
            sweep: Sweep data to write

        Raises:
            WriterError: If no active session or write operation fails
        """
        if self._session_file is None or self._h5_file is None:
            raise WriterError("No active session")

        try:
            sweeps_group = self._h5_file["sweeps"]
            sweep_name = self._unique_child_name(sweeps_group, sweep.name)
            sweep_group = sweeps_group.create_group(sweep_name)

            data_group = sweep_group.create_group("data")
            metadata_group = sweep_group.create_group("metadata")

            # Handle x_label which can be str (1D) or list[str] (2D)
            if isinstance(sweep.x_label, list):
                # 2D sweep: x_data has shape (N, 2)
                for i, label in enumerate(sweep.x_label):
                    self._create_dataset(data_group, label, sweep.x_data[:, i])
            else:
                # 1D sweep
                self._create_dataset(data_group, sweep.x_label, sweep.x_data)
            self._create_dataset(data_group, sweep.y_label, sweep.y_data)

            sweep_group.attrs["x_label"] = sweep.x_label
            sweep_group.attrs["y_label"] = sweep.y_label

            for key, value in sweep.metadata.items():
                try:
                    metadata_group.attrs[key] = value
                except (TypeError, ValueError):
                    metadata_group.attrs[key] = str(value)

            metadata_group.attrs["timestamp"] = to_epoch(sweep.timestamp)
            metadata_group.attrs["session_id"] = sweep.session_id

            logger.debug("Wrote sweep: %s", sweep_name)

        except Exception as e:
            raise WriterError(f"Error writing sweep data: {str(e)}") from e

    def flush(self) -> None:
        """Flush data to disk.

        Raises:
            WriterError: If flush operation fails
        """
        if self._h5_file is None:
            raise WriterError("No active session")

        try:
            self._h5_file.flush()
            logger.debug("Flushed data to disk")
        except Exception as e:
            raise WriterError(f"Error flushing data: {str(e)}") from e

    def _create_dataset(self, group: h5py.Group, name: str, data: np.ndarray) -> None:
        """Create a dataset with the configured compression settings."""
        if self.compression is None:
            group.create_dataset(name, data=data)
        else:
            group.create_dataset(
                name,
                data=data,
                compression=self.compression,
                compression_opts=self.compression_level,
            )

    def _unique_child_name(self, group: h5py.Group, base: str) -> str:
        """Return a unique child name within a group based on base.

        If `base` exists, returns `base_<n>` with the smallest available n.
        """
        if base not in group:
            return base
        i = 0
        while f"{base}_{i}" in group:
            i += 1
        return f"{base}_{i}"



"""Inline plotting backend for Jupyter notebooks.

Uses jupyter_bokeh extension for reliable live updates in all notebook environments.
Install: pip install jupyter_bokeh
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

from bokeh.plotting import output_notebook
from IPython.display import display

try:
    from jupyter_bokeh.widgets import BokehModel  # type: ignore[import-untyped]
except ImportError:
    BokehModel = None

from stanza.plotter.backends.utils import (
    PlotState,
    make_heatmap_plot,
    make_line_plot,
    prepare_heatmap_data,
)

if TYPE_CHECKING:
    pass


class InlineBackend:
    """Display live-updating plots directly in notebook cells."""

    def __init__(self) -> None:
        self._plots: dict[str, PlotState] = {}
        self._displayed: set[str] = set()

    def start(self) -> None:
        """Initialize Bokeh notebook output."""
        if BokehModel is None:
            raise ImportError(
                "jupyter_bokeh is not installed. Install with: pip install jupyter_bokeh"
            )
        output_notebook()

    def stop(self) -> None:
        """Clean up resources."""
        pass

    def create_figure(
        self,
        name: str,
        x_label: str,
        y_label: str,
        plot_type: str = "line",
        z_label: str | None = None,
        cell_size: tuple[float, float] | None = None,
    ) -> None:
        """Create a new plot configuration."""
        if name in self._plots:
            return

        if plot_type == "line":
            plot_state = make_line_plot(name, x_label, y_label)
        elif plot_type == "heatmap":
            plot_state = make_heatmap_plot(name, x_label, y_label, z_label, cell_size)
        else:
            raise ValueError(f"Unknown plot type: {plot_type}")

        self._plots[name] = plot_state

    def stream_data(
        self, name: str, new_data: dict[str, Any], rollover: int | None = None
    ) -> None:
        """Add data to plot and display/update it.

        First call displays the plot, subsequent calls update via ColumnDataSource.
        """
        plot = self._plots.get(name)
        if plot is None:
            return

        if plot.spec.plot_type == "heatmap" and "value" in new_data:
            new_data = prepare_heatmap_data(new_data, plot.source.data, plot.spec)

        # Merge with existing data and apply rollover
        merged_data = {}
        for key, new_vals in new_data.items():
            merged = list(plot.source.data.get(key, [])) + new_vals
            if rollover and len(merged) > rollover:
                merged = merged[-rollover:]
            merged_data[key] = merged

        plot.source.data = merged_data

        if plot.spec.plot_type == "heatmap" and plot.spec.mapper:
            plot.spec.mapper.low = plot.spec.value_min
            plot.spec.mapper.high = plot.spec.value_max

        if name not in self._displayed:
            display(BokehModel(plot.figure))
            self._displayed.add(name)



"""Tests for ServerBackend."""

from unittest.mock import Mock

import pytest
from bokeh.models import ColumnDataSource

from stanza.plotter.backends.server import ServerBackend
from stanza.plotter.backends.utils import (
    PlotSpec,
    PlotState,
    prepare_heatmap_data,
)


@pytest.fixture
def backend():
    """Create ServerBackend instance."""
    return ServerBackend(port=5050, daemon=True)


def test_initialization(backend):
    """Test backend initialization."""
    assert backend.port == 5050
    assert backend.daemon is True
    assert backend._running is False
    assert len(backend._plots) == 0
    assert len(backend._plot_specs) == 0


def test_start_raises_if_not_running():
    """Test that create_figure raises if server not started."""
    backend = ServerBackend(port=5051)

    with pytest.raises(RuntimeError, match="Server not started"):
        backend.create_figure("test", "X", "Y")


def test_create_figure_registers_line_plot():
    """Test that create_figure registers a line plot spec."""
    backend = ServerBackend(port=5052)
    backend._running = True

    backend.create_figure("test_line", "Voltage", "Current", plot_type="line")

    assert "test_line" in backend._plot_specs
    spec = backend._plot_specs["test_line"]
    assert spec.x_label == "Voltage"
    assert spec.y_label == "Current"
    assert spec.plot_type == "line"


def test_create_figure_registers_heatmap():
    """Test that create_figure registers a heatmap spec."""
    backend = ServerBackend(port=5053)
    backend._running = True

    backend.create_figure(
        "test_heatmap",
        "X",
        "Y",
        plot_type="heatmap",
        z_label="Intensity",
        cell_size=(0.1, 0.2),
    )

    assert "test_heatmap" in backend._plot_specs
    spec = backend._plot_specs["test_heatmap"]
    assert spec.plot_type == "heatmap"
    assert spec.z_label == "Intensity"
    assert spec.cell_size == (0.1, 0.2)


def test_stream_data_buffers_before_browser_connects():
    """Test that data is buffered if browser hasn't connected yet."""
    backend = ServerBackend(port=5054)
    backend._running = True

    backend.create_figure("test", "X", "Y")
    backend.stream_data("test", {"x": [1.0, 2.0], "y": [3.0, 4.0]})

    assert "test" in backend._buffer
    assert backend._buffer["test"]["x"] == [1.0, 2.0]
    assert backend._buffer["test"]["y"] == [3.0, 4.0]


def test_stream_data_multiple_times_accumulates_buffer():
    """Test that multiple stream calls accumulate data in buffer."""
    backend = ServerBackend(port=5055)
    backend._running = True

    backend.create_figure("test", "X", "Y")
    backend.stream_data("test", {"x": [1.0], "y": [2.0]})
    backend.stream_data("test", {"x": [3.0], "y": [4.0]})

    assert backend._buffer["test"]["x"] == [1.0, 3.0]
    assert backend._buffer["test"]["y"] == [2.0, 4.0]


def test_stop_stops_running_server():
    """Test that stop() sets running to False."""
    backend = ServerBackend(port=5056)
    backend._running = True
    backend._server = Mock()
    backend._server.io_loop = Mock()

    backend.stop()

    assert backend._running is False
    backend._server.io_loop.stop.assert_called_once()


def test_create_line_plot_with_buffered_data():
    """Test _create_line_plot uses buffered data."""
    backend = ServerBackend(port=5057)
    backend._running = True
    backend._doc = Mock()
    backend._plot_specs["test"] = PlotSpec(
        name="test", x_label="X", y_label="Y", plot_type="line"
    )
    backend._buffer["test"] = {"x": [1.0, 2.0], "y": [3.0, 4.0]}

    backend._create_line_plot("test")

    assert "test" in backend._plots
    assert "test" not in backend._buffer
    backend._doc.add_root.assert_called_once()


def test_create_heatmap_plot_spec():
    """Test _create_heatmap_plot sets up color mapper."""
    backend = ServerBackend(port=5058)
    backend._running = True
    backend._doc = Mock()
    backend._plot_specs["test"] = PlotSpec(
        name="test",
        x_label="X",
        y_label="Y",
        plot_type="heatmap",
        z_label="Value",
        cell_size=(0.5, 0.5),
    )

    backend._create_heatmap_plot("test")

    assert "test" in backend._plots
    plot = backend._plots["test"]
    assert plot.spec.mapper is not None
    assert plot.spec.dx == 0.5
    assert plot.spec.dy == 0.5
    backend._doc.add_root.assert_called_once()


def test_create_plot_raises_on_unknown_type():
    """Test that _create_plot raises on unknown plot type."""
    backend = ServerBackend(port=5059)
    backend._plot_specs["test"] = PlotSpec(
        name="test", x_label="X", y_label="Y", plot_type="unknown"
    )

    with pytest.raises(ValueError, match="Unknown plot type"):
        backend._create_plot("test")


def test_prepare_heatmap_data_updates_color_range():
    """Test that prepare_heatmap_data calculates cell sizes and updates range."""
    backend = ServerBackend(port=5060)
    spec = PlotSpec(name="test", plot_type="heatmap", x_label="X", y_label="Y")
    source = ColumnDataSource(data={"x": [], "y": [], "value": []})
    plot_state = PlotState(source=source, figure=Mock(), spec=spec)
    backend._plots["test"] = plot_state

    data = {"x": [1.0, 2.0], "y": [1.0, 1.0], "value": [5.0, 10.0]}
    result = prepare_heatmap_data(data, plot_state.source.data, plot_state.spec)

    assert "width" in result
    assert "height" in result
    assert plot_state.spec.value_min <= 5.0
    assert plot_state.spec.value_max >= 10.0


def test_start_is_idempotent():
    """Test that calling start multiple times doesn't cause issues."""
    backend = ServerBackend(port=5061)
    backend._running = True

    backend.start()
    assert backend._running is True


def test_create_figure_is_idempotent():
    """Test that creating same figure twice doesn't duplicate."""
    backend = ServerBackend(port=5062)
    backend._running = True

    backend.create_figure("test", "X", "Y")
    backend.create_figure("test", "X", "Y")

    assert len(backend._plot_specs) == 1


def test_create_figure_with_doc_adds_callback():
    """Test that create_figure adds callback when doc is already connected."""
    backend = ServerBackend(port=5063)
    backend._running = True
    backend._doc = Mock()
    backend._doc.add_next_tick_callback = Mock()

    backend.create_figure("test", "X", "Y")

    backend._doc.add_next_tick_callback.assert_called_once()


def test_stream_data_with_doc_adds_callback():
    """Test that streaming data with connected doc adds callback."""
    backend = ServerBackend(port=5064)
    backend._running = True
    backend._doc = Mock()

    spec = PlotSpec(name="test", plot_type="line", x_label="X", y_label="Y")
    source = ColumnDataSource(data={"x": [], "y": []})
    plot_state = PlotState(source=source, figure=Mock(), spec=spec)
    backend._plots["test"] = plot_state

    backend.stream_data("test", {"x": [1.0], "y": [2.0]})

    backend._doc.add_next_tick_callback.assert_called_once()



"""Tests for streaming sweep functionality in LoggerSession.

This module tests the SweepContext API for streaming sweeps with live plotting,
including context manager behavior, data validation, persistence, and integration
with various writers.
"""

import json
import logging
import tempfile
from pathlib import Path
from unittest.mock import MagicMock

import h5py
import numpy as np
import pytest

from stanza.exceptions import LoggerSessionError
from stanza.logger.datatypes import SessionMetadata
from stanza.logger.session import LoggerSession
from stanza.logger.writers.bokeh_writer import BokehLiveWriter
from stanza.logger.writers.hdf5_writer import HDF5Writer
from stanza.logger.writers.jsonl_writer import JSONLWriter


@pytest.fixture
def tmpdir_path():
    with tempfile.TemporaryDirectory() as tmpdir:
        yield Path(tmpdir)


@pytest.fixture
def session_metadata():
    return SessionMetadata(
        session_id="test_session",
        start_time=100.0,
        user="test_user",
    )


@pytest.fixture
def logger_session(tmpdir_path, session_metadata):
    writer = JSONLWriter(tmpdir_path)
    return LoggerSession(
        metadata=session_metadata,
        writer_pool={"jsonl": writer},
        writer_refs=["jsonl"],
        base_dir=tmpdir_path,
    )


class TestSweepContext:
    """Unit tests for SweepContext (streaming sweep handle)."""

    def test_accumulates_data_across_multiple_appends(self, logger_session):
        """Test that multiple append calls accumulate data correctly."""
        logger_session.initialize()

        with logger_session.sweep("signal", "Time", "Amplitude") as sweep:
            sweep.append([0.0, 1.0], [0.5, 0.6])
            sweep.append([2.0], [0.7])
            sweep.append([3.0, 4.0, 5.0], [0.8, 0.9, 1.0])

        assert not sweep._active
        assert len(sweep._x_data) == 6
        assert len(sweep._y_data) == 6

        logger_session.finalize()

    def test_rejects_empty_arrays(self, logger_session):
        """Test that append rejects empty x or y arrays."""
        logger_session.initialize()
        sweep = logger_session.sweep("test", "X", "Y")

        with pytest.raises(ValueError, match="cannot be empty"):
            sweep.append([], [1.0])

        with pytest.raises(ValueError, match="cannot be empty"):
            sweep.append([1.0], [])

        sweep.cancel()
        logger_session.finalize()

    def test_rejects_length_mismatch(self, logger_session):
        """Test that append rejects x and y arrays with different lengths for 1D sweeps."""
        logger_session.initialize()
        sweep = logger_session.sweep("test", "X", "Y")

        # First establish it's a 1D sweep by appending matching lengths
        sweep.append([1.0], [2.0])

        # Now try mismatched lengths - should fail
        with pytest.raises(ValueError, match="Length mismatch"):
            sweep.append([1.0, 2.0], [3.0])

        sweep.cancel()
        logger_session.finalize()

    def test_enforces_1d_arrays_only(self, logger_session):
        """Test that multidimensional numpy arrays are rejected."""
        logger_session.initialize()
        sweep = logger_session.sweep("test", "X", "Y")

        with pytest.raises(ValueError, match="must be 1D arrays"):
            sweep.append(np.array([[1, 2], [3, 4]]), [1.0, 2.0])

        with pytest.raises(ValueError, match="must be 1D arrays"):
            sweep.append([1.0, 2.0], np.array([[1, 2], [3, 4]]))

        sweep.cancel()
        logger_session.finalize()

    def test_cancel_prevents_persistence(self, tmpdir_path, logger_session):
        """Test that cancel() prevents data from being written to file."""
        logger_session.initialize()

        sweep = logger_session.sweep("test", "X", "Y")
        sweep.append([1.0, 2.0], [3.0, 4.0])
        sweep.cancel()

        assert not sweep._active
        logger_session.finalize()

        sweep_file = tmpdir_path / "sweep.jsonl"
        assert not sweep_file.exists() or sweep_file.stat().st_size == 0

    def test_end_writes_to_file(self, tmpdir_path, logger_session):
        """Test that end() writes accumulated data to file."""
        logger_session.initialize()

        sweep = logger_session.sweep("test_sweep", "X", "Y")
        sweep.append([1.0, 2.0], [3.0, 4.0])
        sweep.end()

        assert not sweep._active
        logger_session.finalize()

        sweep_file = tmpdir_path / "sweep.jsonl"
        assert sweep_file.exists()
        assert sweep_file.stat().st_size > 0

    def test_context_exit_normal_calls_end(self, tmpdir_path, logger_session):
        """Test that normal context exit calls end()."""
        logger_session.initialize()

        with logger_session.sweep("signal", "Time", "Amplitude") as sweep:
            sweep.append([1.0], [2.0])

        assert not sweep._active
        logger_session.finalize()

        sweep_file = tmpdir_path / "sweep.jsonl"
        assert sweep_file.exists()

    def test_context_exit_with_exception_calls_cancel(
        self, tmpdir_path, logger_session
    ):
        """Test that exception in context calls cancel() and doesn't persist."""
        logger_session.initialize()

        with pytest.raises(RuntimeError):
            with logger_session.sweep("signal", "Time", "Amplitude") as sweep:
                sweep.append([1.0], [2.0])
                raise RuntimeError("Test exception")

        assert not sweep._active
        logger_session.finalize()

        sweep_file = tmpdir_path / "sweep.jsonl"
        assert not sweep_file.exists() or sweep_file.stat().st_size == 0

    def test_append_after_end_raises_error(self, logger_session):
        """Test that append after end() raises error."""
        logger_session.initialize()

        sweep = logger_session.sweep("test", "X", "Y")
        sweep.append([1.0], [2.0])
        sweep.end()

        with pytest.raises(ValueError, match="no longer active"):
            sweep.append([3.0], [4.0])

        logger_session.finalize()

    def test_append_after_cancel_raises_error(self, logger_session):
        """Test that append after cancel() raises error."""
        logger_session.initialize()

        sweep = logger_session.sweep("test", "X", "Y")
        sweep.append([1.0], [2.0])
        sweep.cancel()

        with pytest.raises(ValueError, match="no longer active"):
            sweep.append([3.0], [4.0])

        logger_session.finalize()

    def test_end_with_no_data_does_nothing(self, logger_session, caplog):
        """Test that end() on empty sweep logs debug and doesn't write."""
        logger_session.initialize()

        with caplog.at_level(logging.DEBUG):
            sweep = logger_session.sweep("empty", "X", "Y")
            sweep.end()

        assert any(
            "empty, skipping write" in record.message for record in caplog.records
        )
        logger_session.finalize()

    def test_double_end_is_safe(self, logger_session):
        """Test that calling end() multiple times is safe (idempotent)."""
        logger_session.initialize()

        sweep = logger_session.sweep("test", "X", "Y")
        sweep.append([1.0], [2.0])
        sweep.end()
        sweep.end()
        sweep.end()

        logger_session.finalize()

    def test_converts_lists_to_numpy_arrays(self, logger_session):
        """Test that append accepts and converts lists to numpy arrays."""
        logger_session.initialize()

        sweep = logger_session.sweep("test", "X", "Y")
        sweep.append([1.0, 2.0, 3.0], [4.0, 5.0, 6.0])

        assert len(sweep._x_data) == 3
        assert len(sweep._y_data) == 3

        sweep.end()
        logger_session.finalize()

    def test_metadata_is_copied_defensively(self, logger_session):
        """Test that metadata is copied to prevent external mutation."""
        logger_session.initialize()

        original_metadata = {"key": "value"}
        sweep = logger_session.sweep("test", "X", "Y", metadata=original_metadata)

        original_metadata["key"] = "changed"
        original_metadata["new_key"] = "new_value"

        assert sweep.metadata == {"key": "value"}

        sweep.cancel()
        logger_session.finalize()


class TestStreamingSweepIntegration:
    """Integration tests for streaming sweep functionality with session."""

    def test_session_sweep_creates_context(self, logger_session):
        """Test that session.sweep() creates and registers a SweepContext."""
        logger_session.initialize()

        sweep = logger_session.sweep("signal", "Time", "Amplitude")

        assert sweep.name == "signal"
        assert sweep.x_label == "Time"
        assert sweep.y_label == "Amplitude"
        assert "signal" in logger_session._active_sweeps

        sweep.cancel()
        logger_session.finalize()

    def test_sweep_creates_multiple_contexts(self, logger_session):
        """Test that sweep() can create multiple independent contexts."""
        logger_session.initialize()

        sweep1 = logger_session.sweep("test1", "X", "Y")
        sweep2 = logger_session.sweep("test2", "X", "Y")

        assert type(sweep1) is type(sweep2)
        assert "test1" in logger_session._active_sweeps
        assert "test2" in logger_session._active_sweeps

        sweep1.cancel()
        sweep2.cancel()
        logger_session.finalize()

    def test_rejects_duplicate_active_sweep_name(self, logger_session):
        """Test that starting a sweep with an active name raises error."""
        logger_session.initialize()

        sweep1 = logger_session.sweep("duplicate", "X", "Y")

        with pytest.raises(LoggerSessionError, match="already active"):
            logger_session.sweep("duplicate", "X", "Y")

        sweep1.end()
        sweep2 = logger_session.sweep("duplicate", "X", "Y")
        sweep2.cancel()

        logger_session.finalize()

    def test_sweep_before_initialize_raises_error(self, logger_session):
        """Test that creating sweep before initialization raises error."""
        with pytest.raises(LoggerSessionError, match="not initialized"):
            logger_session.sweep("test", "X", "Y")

    def test_live_streaming_calls_bokeh_writer(self, tmpdir_path, session_metadata):
        """Test that append() streams to BokehLiveWriter on each call."""
        mock_bokeh = MagicMock(spec=BokehLiveWriter)
        jsonl_writer = JSONLWriter(tmpdir_path)

        session = LoggerSession(
            metadata=session_metadata,
            writer_pool={"bokeh": mock_bokeh, "jsonl": jsonl_writer},
            writer_refs=["bokeh", "jsonl"],
            base_dir=tmpdir_path,
        )

        session.initialize()

        with session.sweep("signal", "Time", "Amplitude") as sweep:
            sweep.append([1.0], [2.0])
            sweep.append([3.0], [4.0])

        assert mock_bokeh.write_sweep.call_count == 2

        session.finalize()

    def test_persistence_skips_bokeh_writer(self, tmpdir_path, session_metadata):
        """Test that end() writes to file writers but skips BokehLiveWriter."""
        mock_bokeh = MagicMock(spec=BokehLiveWriter)
        jsonl_writer = JSONLWriter(tmpdir_path)

        session = LoggerSession(
            metadata=session_metadata,
            writer_pool={"bokeh": mock_bokeh, "jsonl": jsonl_writer},
            writer_refs=["bokeh", "jsonl"],
            base_dir=tmpdir_path,
        )

        session.initialize()

        with session.sweep("signal", "Time", "Amplitude") as sweep:
            sweep.append([1.0, 2.0], [3.0, 4.0])

        assert mock_bokeh.write_sweep.call_count == 1

        session.finalize()

        sweep_file = tmpdir_path / "sweep.jsonl"
        assert sweep_file.exists()
        assert sweep_file.stat().st_size > 0

    def test_writes_exactly_one_record_to_file(self, tmpdir_path, logger_session):
        """Test that streaming sweep writes exactly one record to file."""
        logger_session.initialize()

        with logger_session.sweep("multi_append", "X", "Y") as sweep:
            for i in range(10):
                sweep.append([float(i)], [float(i * 2)])

        logger_session.finalize()

        sweep_file = tmpdir_path / "sweep.jsonl"
        with open(sweep_file) as f:
            lines = f.readlines()

        assert len(lines) == 1

        record = json.loads(lines[0])
        assert record["name"] == "multi_append"
        assert len(record["x_data"]) == 10
        assert len(record["y_data"]) == 10

    def test_finalize_auto_ends_active_sweeps(
        self, tmpdir_path, logger_session, caplog
    ):
        """Test that finalize() automatically ends active sweeps with warning."""
        logger_session.initialize()

        sweep1 = logger_session.sweep("active1", "X", "Y")
        sweep1.append([1.0], [2.0])

        sweep2 = logger_session.sweep("active2", "X", "Y")
        sweep2.append([3.0], [4.0])

        with caplog.at_level(logging.WARNING):
            logger_session.finalize()

        assert any("Auto-completing" in record.message for record in caplog.records)
        assert any("active1" in record.message for record in caplog.records)

        assert not sweep1._active
        assert not sweep2._active

        sweep_file = tmpdir_path / "sweep.jsonl"
        assert sweep_file.exists()

    def test_coexistence_with_log_sweep(self, tmpdir_path, logger_session):
        """Test that streaming sweeps and log_sweep() can coexist without duplicates."""
        logger_session.initialize()

        logger_session.log_sweep(
            "complete_sweep", [1.0, 2.0, 3.0], [4.0, 5.0, 6.0], "X", "Y"
        )

        with logger_session.sweep("streaming_sweep", "X", "Y") as sweep:
            sweep.append([7.0, 8.0], [9.0, 10.0])

        logger_session.finalize()

        sweep_file = tmpdir_path / "sweep.jsonl"
        with open(sweep_file) as f:
            lines = f.readlines()

        assert len(lines) == 2

        records = [json.loads(line) for line in lines]
        names = {r["name"] for r in records}
        assert names == {"complete_sweep", "streaming_sweep"}

    def test_streaming_sweep_with_hdf5_writer(self, tmpdir_path, session_metadata):
        """Test that streaming sweeps work with HDF5Writer."""
        hdf5_writer = HDF5Writer(tmpdir_path)

        session = LoggerSession(
            metadata=session_metadata,
            writer_pool={"hdf5": hdf5_writer},
            writer_refs=["hdf5"],
            base_dir=tmpdir_path,
        )

        session.initialize()

        with session.sweep("hdf5_sweep", "Voltage", "Current") as sweep:
            sweep.append([0.0, 0.1], [0.0, 1e-6])
            sweep.append([0.2, 0.3], [2e-6, 3e-6])

        session.finalize()

        hdf5_file = tmpdir_path / f"{session_metadata.session_id}.h5"
        assert hdf5_file.exists()

        with h5py.File(hdf5_file, "r") as f:
            assert "sweeps" in f
            assert "hdf5_sweep" in f["sweeps"]

    def test_multiple_writers_receive_complete_sweep(
        self, tmpdir_path, session_metadata
    ):
        """Test that all file writers receive the complete sweep on end()."""
        jsonl_dir = tmpdir_path / "jsonl"
        hdf5_dir = tmpdir_path / "hdf5"
        jsonl_dir.mkdir()
        hdf5_dir.mkdir()

        jsonl_writer = JSONLWriter(jsonl_dir)
        hdf5_writer = HDF5Writer(hdf5_dir)

        session = LoggerSession(
            metadata=session_metadata,
            writer_pool={"jsonl": jsonl_writer, "hdf5": hdf5_writer},
            writer_refs=["jsonl", "hdf5"],
            base_dir=tmpdir_path,
        )

        session.initialize()

        with session.sweep("multi_writer", "X", "Y") as sweep:
            sweep.append([1.0, 2.0, 3.0], [4.0, 5.0, 6.0])

        session.finalize()

        jsonl_file = jsonl_dir / "sweep.jsonl"
        assert jsonl_file.exists()

        hdf5_file = hdf5_dir / f"{session_metadata.session_id}.h5"
        assert hdf5_file.exists()

        with h5py.File(hdf5_file, "r") as f:
            assert "multi_writer" in f["sweeps"]

    def test_streaming_sweep_preserves_metadata(self, tmpdir_path, logger_session):
        """Test that user metadata is preserved in the persisted sweep."""
        logger_session.initialize()

        custom_metadata = {"temperature": 300, "device": "QD1", "notes": "test run"}

        with logger_session.sweep(
            "meta_sweep", "Voltage", "Current", metadata=custom_metadata
        ) as sweep:
            sweep.append([1.0], [2.0])

        logger_session.finalize()

        sweep_file = tmpdir_path / "sweep.jsonl"
        with open(sweep_file) as f:
            record = json.loads(f.readline())

        assert record["metadata"] == custom_metadata

    def test_routine_name_attached_to_streaming_sweep(
        self, tmpdir_path, session_metadata
    ):
        """Test that routine name is attached to streaming sweep."""
        session_metadata.routine_name = "test_routine"
        jsonl_writer = JSONLWriter(tmpdir_path)

        session = LoggerSession(
            metadata=session_metadata,
            writer_pool={"jsonl": jsonl_writer},
            writer_refs=["jsonl"],
            base_dir=tmpdir_path,
        )

        session.initialize()

        with session.sweep("routine_sweep", "X", "Y") as sweep:
            sweep.append([1.0], [2.0])

        session.finalize()

        sweep_file = tmpdir_path / "sweep.jsonl"
        with open(sweep_file) as f:
            record = json.loads(f.readline())

        assert record["routine_name"] == "test_routine"

    def test_live_writer_error_doesnt_stop_persistence(
        self, tmpdir_path, session_metadata, caplog
    ):
        """Test that live writer errors don't prevent file persistence."""
        mock_bokeh = MagicMock(spec=BokehLiveWriter)
        mock_bokeh.write_sweep.side_effect = RuntimeError("Bokeh plot failed")

        jsonl_writer = JSONLWriter(tmpdir_path)

        session = LoggerSession(
            metadata=session_metadata,
            writer_pool={"bokeh": mock_bokeh, "jsonl": jsonl_writer},
            writer_refs=["bokeh", "jsonl"],
            base_dir=tmpdir_path,
        )

        session.initialize()

        with caplog.at_level(logging.WARNING):
            with session.sweep("resilient", "X", "Y") as sweep:
                sweep.append([1.0], [2.0])

        assert any(
            "Live plot stream failed" in record.message for record in caplog.records
        )

        session.finalize()

        sweep_file = tmpdir_path / "sweep.jsonl"
        assert sweep_file.exists()

    def test_file_writer_error_during_persistence_raises(
        self, tmpdir_path, session_metadata
    ):
        """Test that file writer errors during persistence raise LoggerSessionError."""

        class FailingWriter(JSONLWriter):
            def write_sweep(self, sweep):
                raise RuntimeError("File write failed")

        writer = FailingWriter(tmpdir_path)

        session = LoggerSession(
            metadata=session_metadata,
            writer_pool={"jsonl": writer},
            writer_refs=["jsonl"],
            base_dir=tmpdir_path,
        )

        session.initialize()

        with pytest.raises(LoggerSessionError, match="Sweep write failed"):
            with session.sweep("failing", "X", "Y") as sweep:
                sweep.append([1.0], [2.0])

        session._active = False  # Force inactive to allow cleanup

    def test_no_live_writers_doesnt_error(self, tmpdir_path, logger_session):
        """Test that streaming works when no live writers are configured."""
        logger_session.initialize()

        with logger_session.sweep("no_live", "X", "Y") as sweep:
            sweep.append([1.0, 2.0], [3.0, 4.0])

        logger_session.finalize()

        sweep_file = tmpdir_path / "sweep.jsonl"
        assert sweep_file.exists()

    def test_session_repr_includes_active_sweeps(self, logger_session):
        """Test that session repr includes count of active sweeps."""
        logger_session.initialize()

        repr_before = repr(logger_session)
        assert "active_sweeps=0" in repr_before

        sweep1 = logger_session.sweep("s1", "X", "Y")
        sweep2 = logger_session.sweep("s2", "X", "Y")

        repr_during = repr(logger_session)
        assert "active_sweeps=2" in repr_during

        sweep1.end()
        sweep2.end()

        repr_after = repr(logger_session)
        assert "active_sweeps=0" in repr_after

        logger_session.finalize()



from stanza.base.protocols import (
    BreakoutBoxInstrument,
    ControlInstrument,
    MeasurementInstrument,
)
from tests.conftest import (
    MockBreakoutBoxInstrument,
    MockControlInstrument,
    MockMeasurementInstrument,
)


class TestControlInstrumentProtocol:
    def test_protocol_implementation(self):
        mock_instrument = MockControlInstrument()

        assert isinstance(mock_instrument, ControlInstrument)

        mock_instrument.set_voltage("gate1", 1.5)
        voltage = mock_instrument.get_voltage("gate1")

        assert voltage == 1.5

    def test_protocol_methods_exist(self):
        mock_instrument = MockControlInstrument()

        assert hasattr(mock_instrument, "set_voltage")
        assert hasattr(mock_instrument, "get_voltage")
        assert hasattr(mock_instrument, "get_slew_rate")


class TestMeasurementInstrumentProtocol:
    def test_protocol_implementation(self):
        mock_instrument = MockMeasurementInstrument()

        assert isinstance(mock_instrument, MeasurementInstrument)

        mock_instrument.measurements["sense1"] = 1e-6
        current = mock_instrument.measure("sense1")

        assert current == 1e-6

    def test_protocol_methods_exist(self):
        mock_instrument = MockMeasurementInstrument()

        assert hasattr(mock_instrument, "measure")


class TestBreakoutBoxInstrumentProtocol:
    def test_protocol_implementation(self):
        mock_instrument = MockBreakoutBoxInstrument()
        assert isinstance(mock_instrument, BreakoutBoxInstrument)

    def test_protocol_methods_exist(self):
        mock_instrument = MockBreakoutBoxInstrument()
        assert hasattr(mock_instrument, "get_grounded")
        assert hasattr(mock_instrument, "set_grounded")
        assert hasattr(mock_instrument, "get_ungrounded")
        assert hasattr(mock_instrument, "set_ungrounded")
        assert hasattr(mock_instrument, "get_connected")
        assert hasattr(mock_instrument, "set_connected")
        assert hasattr(mock_instrument, "get_disconnected")
        assert hasattr(mock_instrument, "set_disconnected")

    def test_grounding_functionality(self):
        mock_instrument = MockBreakoutBoxInstrument()

        assert mock_instrument.get_ungrounded("ch1") is True
        assert mock_instrument.get_grounded("ch1") is False

        mock_instrument.set_grounded("ch1")
        assert mock_instrument.get_grounded("ch1") is True
        assert mock_instrument.get_ungrounded("ch1") is False

        mock_instrument.set_ungrounded("ch1")
        assert mock_instrument.get_ungrounded("ch1") is True
        assert mock_instrument.get_grounded("ch1") is False

    def test_connection_functionality(self):
        mock_instrument = MockBreakoutBoxInstrument()

        assert mock_instrument.get_disconnected("ch1", 5) is True
        assert mock_instrument.get_connected("ch1", 5) is False

        mock_instrument.set_connected("ch1", 5)
        assert mock_instrument.get_connected("ch1", 5) is True
        assert mock_instrument.get_disconnected("ch1", 5) is False
        assert ("ch1", 5) in mock_instrument.connected_calls

        mock_instrument.set_disconnected("ch1", 5)
        assert ("ch1", 5) in mock_instrument.disconnected_calls



"""Tests for BokehLiveWriter."""

import getpass

import numpy as np
import pytest

from stanza.logger.datatypes import MeasurementData, SessionMetadata, SweepData
from stanza.logger.writers.bokeh_writer import BokehLiveWriter


class MockBackend:
    """Mock backend for testing."""

    def __init__(self) -> None:
        self.figures_created: list[dict[str, str]] = []
        self.push_count = 0
        self.streamed_data: dict[str, list[dict]] = {}

    def create_figure(
        self,
        name: str,
        x_label: str,
        y_label: str,
        plot_type: str = "line",
        z_label: str | None = None,
        cell_size: tuple[float, float] | None = None,
    ) -> object:
        """Mock create_figure."""
        self.figures_created.append(
            {
                "name": name,
                "x_label": x_label,
                "y_label": y_label,
                "plot_type": plot_type,
            }
        )
        fig = type("MockFig", (), {"line": lambda *args, **kwargs: None})()
        return fig

    def stream_data(self, name: str, new_data: dict, rollover: int) -> None:
        """Mock stream_data."""
        if name not in self.streamed_data:
            self.streamed_data[name] = []
        self.streamed_data[name].append({"data": new_data, "rollover": rollover})

    def push_updates(self) -> None:
        """Mock push_updates."""
        self.push_count += 1


@pytest.fixture
def mock_backend() -> MockBackend:
    """Provide a mock backend."""
    return MockBackend()


@pytest.fixture
def writer(mock_backend: MockBackend) -> BokehLiveWriter:
    """Provide a BokehLiveWriter with mock backend."""
    return BokehLiveWriter(backend=mock_backend, max_points=100)


@pytest.fixture
def session_metadata() -> SessionMetadata:
    """Provide session metadata."""
    return SessionMetadata(
        session_id="test_session",
        routine_name="test_routine",
        start_time=1234567890.0,
        user=getpass.getuser(),
    )


def test_initialization(mock_backend: MockBackend) -> None:
    """Test writer initialization."""
    writer = BokehLiveWriter(backend=mock_backend, max_points=500)

    assert writer.backend is mock_backend
    assert writer.max_points == 500
    assert len(writer._plots) == 0


def test_initialize_session(
    writer: BokehLiveWriter, session_metadata: SessionMetadata
) -> None:
    """Test session initialization."""
    writer.initialize_session(session_metadata)

    assert writer._initialized is True


def test_write_measurement_no_op(
    writer: BokehLiveWriter, session_metadata: SessionMetadata
) -> None:
    """Test that write_measurement is a no-op."""
    writer.initialize_session(session_metadata)

    measurement = MeasurementData(
        name="test",
        data={"current": 1e-9},
        metadata={},
        timestamp=1234567890.0,
        session_id="test_session",
        routine_name="test_routine",
    )

    writer.write_measurement(measurement)


def test_write_sweep_creates_figure(
    writer: BokehLiveWriter,
    mock_backend: MockBackend,
    session_metadata: SessionMetadata,
) -> None:
    """Test that first sweep creates a figure."""
    writer.initialize_session(session_metadata)

    sweep = SweepData(
        name="G1_sweep",
        x_data=np.array([1.0, 2.0, 3.0]),
        y_data=np.array([0.1, 0.2, 0.3]),
        x_label="Voltage (V)",
        y_label="Current (A)",
        metadata={},
        timestamp=1234567890.0,
        session_id="test_session",
    )

    writer.write_sweep(sweep)

    assert len(mock_backend.figures_created) == 1
    assert mock_backend.figures_created[0]["name"] == "G1_sweep"
    assert "G1_sweep" in writer._plots


def test_flush_calls_backend_push(
    writer: BokehLiveWriter,
    mock_backend: MockBackend,
    session_metadata: SessionMetadata,
) -> None:
    """Test that flush calls backend.push_updates()."""
    writer.initialize_session(session_metadata)

    assert mock_backend.push_count == 0

    writer.flush()
    assert mock_backend.push_count == 1

    writer.flush()
    assert mock_backend.push_count == 2


def test_finalize_calls_flush(
    writer: BokehLiveWriter,
    mock_backend: MockBackend,
    session_metadata: SessionMetadata,
) -> None:
    """Test that finalize calls flush."""
    writer.initialize_session(session_metadata)

    assert mock_backend.push_count == 0

    writer.finalize_session(session_metadata)

    assert mock_backend.push_count == 1



"""Live plotting for data logging.

Two backends available:
- server: Plots in browser window (works everywhere)
- inline: Plots in notebook cells (requires jupyter_bokeh)
"""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Literal

from stanza.plotter.backends.inline import InlineBackend
from stanza.plotter.backends.server import ServerBackend

if TYPE_CHECKING:
    from stanza.logger.data_logger import DataLogger

logger = logging.getLogger(__name__)

_active_servers: dict[int, ServerBackend] = {}


def enable_live_plotting(
    data_logger: DataLogger,
    backend: Literal["server", "inline"] = "server",
    port: int = 5006,
) -> ServerBackend | InlineBackend:
    """Enable live plotting for a data logger.

    Args:
        data_logger: DataLogger instance
        backend: "server" (browser) or "inline" (notebook)
        port: Server port (server backend only)

    Returns:
        Backend instance

    Example (server):
        >>> backend = enable_live_plotting(logger, backend="server", port=5006)
        >>> # Open http://localhost:5006 in browser

    Example (inline):
        >>> backend = enable_live_plotting(logger, backend="inline")
        >>> # Plots appear in notebook cells
    """
    bokeh_backend: ServerBackend | InlineBackend

    if backend == "server":
        if port in _active_servers and _active_servers[port]._running:
            bokeh_backend = _active_servers[port]
            logger.info(f"Reusing existing Bokeh Server on port {port}")
            bokeh_backend.reset()
        else:
            if port in _active_servers:
                del _active_servers[port]

            bokeh_backend = ServerBackend(port=port)
            bokeh_backend.start()
            _active_servers[port] = bokeh_backend
            logger.info(f"Bokeh Server started: http://localhost:{port}")

    elif backend == "inline":
        bokeh_backend = InlineBackend()
        bokeh_backend.start()
        logger.info("Bokeh Inline plotting enabled")

    else:
        raise ValueError(f"Unknown backend: {backend}")

    data_logger._bokeh_backend = bokeh_backend
    return bokeh_backend


__all__ = ["enable_live_plotting"]



"""Tests for InlineBackend."""

from unittest.mock import Mock, patch

import pytest
from bokeh.models import ColumnDataSource, LinearColorMapper

from stanza.plotter.backends.inline import InlineBackend
from stanza.plotter.backends.utils import PlotSpec, PlotState


@pytest.fixture
def mock_bokeh():
    """Mock bokeh and jupyter imports."""
    with (
        patch("stanza.plotter.backends.inline.output_notebook") as mock_output,
        patch("stanza.plotter.backends.inline.display") as mock_display,
        patch("stanza.plotter.backends.inline.make_line_plot") as mock_make_line,
        patch("stanza.plotter.backends.inline.make_heatmap_plot") as mock_make_heatmap,
        patch("stanza.plotter.backends.inline.BokehModel") as mock_bokeh_model,
    ):
        mock_fig = Mock()
        mock_source = Mock(spec=ColumnDataSource)
        mock_source.data = {}
        mock_mapper = Mock(spec=LinearColorMapper)

        line_spec = PlotSpec(name="test", plot_type="line", x_label="X", y_label="Y")
        heatmap_spec = PlotSpec(
            name="test",
            plot_type="heatmap",
            x_label="X",
            y_label="Y",
            z_label="Value",
            cell_size=(0.5, 0.5),
            mapper=mock_mapper,
            dx=0.5,
            dy=0.5,
        )

        line_state = PlotState(source=mock_source, figure=mock_fig, spec=line_spec)
        heatmap_state = PlotState(
            source=mock_source, figure=mock_fig, spec=heatmap_spec
        )

        mock_make_line.return_value = line_state
        mock_make_heatmap.return_value = heatmap_state

        yield {
            "output_notebook": mock_output,
            "display": mock_display,
            "make_line_plot": mock_make_line,
            "make_heatmap_plot": mock_make_heatmap,
            "BokehModel": mock_bokeh_model,
        }


def test_initialization(mock_bokeh):
    """Test backend initialization."""
    backend = InlineBackend()

    assert len(backend._plots) == 0
    assert len(backend._displayed) == 0


def test_start_calls_output_notebook(mock_bokeh):
    """Test that start calls output_notebook."""
    backend = InlineBackend()
    backend.start()

    mock_bokeh["output_notebook"].assert_called_once()


def test_stop_cleanup(mock_bokeh):
    """Test that stop can be called."""
    backend = InlineBackend()
    backend.stop()


def test_create_line_plot(mock_bokeh):
    """Test creating a line plot."""
    backend = InlineBackend()

    backend.create_figure("test_line", "Voltage", "Current", plot_type="line")

    assert "test_line" in backend._plots
    assert backend._plots["test_line"].spec.plot_type == "line"


def test_create_heatmap_plot(mock_bokeh):
    """Test creating a heatmap plot."""
    backend = InlineBackend()

    backend.create_figure(
        "test_heatmap",
        "X",
        "Y",
        plot_type="heatmap",
        z_label="Intensity",
        cell_size=(0.5, 0.5),
    )

    assert "test_heatmap" in backend._plots
    plot = backend._plots["test_heatmap"]
    assert plot.spec.plot_type == "heatmap"
    assert plot.spec.dx == 0.5
    assert plot.spec.dy == 0.5
    assert plot.spec.mapper is not None


def test_create_unknown_plot_type_raises(mock_bokeh):
    """Test that unknown plot type raises ValueError."""
    backend = InlineBackend()

    with pytest.raises(ValueError, match="Unknown plot type"):
        backend.create_figure("test", "X", "Y", plot_type="invalid")


def test_stream_data_line_plot_displays_on_first_call(mock_bokeh):
    """Test that line plot is displayed on first stream."""
    backend = InlineBackend()
    backend.create_figure("test", "X", "Y", plot_type="line")

    backend.stream_data("test", {"x": [1.0, 2.0], "y": [3.0, 4.0]})

    assert "test" in backend._displayed
    mock_bokeh["display"].assert_called_once()


def test_stream_data_heatmap_displays_on_first_call(mock_bokeh):
    """Test that heatmap is displayed on first stream."""
    backend = InlineBackend()
    backend.create_figure("test", "X", "Y", plot_type="heatmap", cell_size=(0.1, 0.1))

    backend.stream_data(
        "test", {"x": [1.0, 2.0], "y": [1.0, 2.0], "value": [5.0, 10.0]}
    )

    assert "test" in backend._displayed
    mock_bokeh["display"].assert_called_once()


def test_stream_data_heatmap_updates_color_mapper(mock_bokeh):
    """Test that streaming heatmap data updates color mapper range."""
    backend = InlineBackend()
    backend.create_figure("test", "X", "Y", plot_type="heatmap")

    backend.stream_data("test", {"x": [1.0], "y": [1.0], "value": [5.0]})
    plot = backend._plots["test"]

    assert plot.spec.value_min <= 5.0
    assert plot.spec.value_max >= 5.0


def test_stream_data_with_rollover(mock_bokeh):
    """Test that rollover limits data size."""
    backend = InlineBackend()
    backend.create_figure("test", "X", "Y", plot_type="line")

    backend.stream_data(
        "test", {"x": [1.0, 2.0, 3.0], "y": [4.0, 5.0, 6.0]}, rollover=2
    )

    plot = backend._plots["test"]
    assert len(plot.source.data["x"]) == 2
    assert plot.source.data["x"] == [2.0, 3.0]


def test_stream_data_ignores_unknown_plot(mock_bokeh):
    """Test that streaming to non-existent plot is a no-op."""
    backend = InlineBackend()

    backend.stream_data("nonexistent", {"x": [1.0], "y": [2.0]})


def test_stream_data_accumulates_multiple_calls(mock_bokeh):
    """Test that multiple stream calls accumulate data."""
    backend = InlineBackend()
    backend.create_figure("test", "X", "Y", plot_type="line")

    backend.stream_data("test", {"x": [1.0], "y": [2.0]})
    backend.stream_data("test", {"x": [3.0], "y": [4.0]})

    plot = backend._plots["test"]
    assert plot.source.data["x"] == [1.0, 3.0]
    assert plot.source.data["y"] == [2.0, 4.0]


def test_create_figure_idempotent(mock_bokeh):
    """Test that creating same figure twice doesn't duplicate."""
    backend = InlineBackend()

    backend.create_figure("test", "X", "Y")
    backend.create_figure("test", "X", "Y")

    assert len(backend._plots) == 1



"""Tests for plotter enable_live_plotting function."""

from unittest.mock import Mock, patch

import pytest

from stanza.plotter import enable_live_plotting


@pytest.fixture
def mock_data_logger():
    """Mock DataLogger instance."""
    return Mock()


def test_enable_server_backend_reuses_existing_port(mock_data_logger):
    """Test that server backend reuses existing server on same port."""
    import time

    backend1 = enable_live_plotting(mock_data_logger, backend="server", port=5006)
    assert backend1.port == 5006

    mock_data_logger2 = Mock()
    backend2 = enable_live_plotting(mock_data_logger2, backend="server", port=5006)

    assert backend2 is backend1
    assert backend2.port == 5006
    assert mock_data_logger2._bokeh_backend is backend2

    backend1.stop()
    time.sleep(0.1)


def test_enable_inline_backend(mock_data_logger):
    """Test enabling inline backend."""
    with patch("stanza.plotter.InlineBackend") as MockInlineBackend:
        mock_backend = Mock()
        MockInlineBackend.return_value = mock_backend

        backend = enable_live_plotting(mock_data_logger, backend="inline")

        mock_backend.start.assert_called_once()
        assert backend is mock_backend
        assert mock_data_logger._bokeh_backend is mock_backend


def test_enable_unknown_backend_raises(mock_data_logger):
    """Test that unknown backend raises ValueError."""
    with pytest.raises(ValueError, match="Unknown backend"):
        enable_live_plotting(mock_data_logger, backend="invalid")



"""Bokeh backend type alias."""

from __future__ import annotations

from typing import Union

from stanza.plotter.backends.inline import InlineBackend
from stanza.plotter.backends.server import ServerBackend

BokehBackend = Union["ServerBackend", "InlineBackend"]



import importlib.resources
from importlib.resources import as_file

import pytest

from stanza.base.channels import ChannelConfig
from stanza.drivers.qswitch import QSwitch, QSwitchChannel
from stanza.models import BaseInstrumentConfig, GateType, InstrumentType, PadType


@pytest.fixture
def qswitch_sim():
    instrument_config = BaseInstrumentConfig(
        name="qswitch_sim",
        type=InstrumentType.BREAKOUT_BOX,
        serial_addr="192.168.1.100",
        port=5025,
    )

    channel_configs = {
        "ch1": ChannelConfig(
            name="ch1",
            voltage_range=(0.0, 5.0),
            pad_type=PadType.GATE,
            electrode_type=GateType.PLUNGER,
            breakout_channel=1,
        ),
    }

    with as_file(
        importlib.resources.files(__package__).joinpath("qswitch_pyvisa_sim.yaml")
    ) as sim_file:
        return QSwitch(
            instrument_config=instrument_config,
            channel_configs=channel_configs,
            is_simulation=True,
            sim_file=str(sim_file),
        )


class TestQSwitchPyVisaSim:
    def test_idn_query(self, qswitch_sim):
        idn = qswitch_sim.driver.query("*IDN?")
        assert "QDevil,QSwitch" in idn
        assert "QSWSIM1234" in idn

    def test_channel_initialization(self, qswitch_sim):
        assert "ch1" in qswitch_sim.channels
        assert isinstance(qswitch_sim.channels["ch1"], QSwitchChannel)

    def test_parameter_access(self, qswitch_sim):
        channel = qswitch_sim.channels["ch1"]
        assert channel.get_parameter("connect_relay") is not None
        assert channel.get_parameter("disconnect_relay") is not None

    def test_get_grounded(self, qswitch_sim):
        assert qswitch_sim.get_grounded("ch1") is True

    def test_get_ungrounded(self, qswitch_sim):
        assert qswitch_sim.get_ungrounded("ch1") is False

    def test_get_connected(self, qswitch_sim):
        assert qswitch_sim.get_connected("ch1", 0) is True
        assert qswitch_sim.get_connected("ch1", 1) is False
        assert qswitch_sim.get_connected("ch1", 5) is False

    def test_get_disconnected(self, qswitch_sim):
        assert qswitch_sim.get_disconnected("ch1", 0) is False
        assert qswitch_sim.get_disconnected("ch1", 1) is True
        assert qswitch_sim.get_disconnected("ch1", 5) is True

    def test_set_operations(self, qswitch_sim):
        qswitch_sim.set_grounded("ch1")
        qswitch_sim.set_ungrounded("ch1")
        qswitch_sim.set_connected("ch1", 3)
        qswitch_sim.set_disconnected("ch1", 5)

    def test_close(self, qswitch_sim):
        qswitch_sim.close()



from __future__ import annotations

import logging
import time
from pathlib import Path
from typing import Any

import numpy as np

from stanza.exceptions import LoggerSessionError
from stanza.logger.datatypes import MeasurementData, SessionMetadata, SweepData
from stanza.logger.writers.base import AbstractDataWriter
from stanza.logger.writers.bokeh_writer import BokehLiveWriter

logger = logging.getLogger(__name__)


class SweepContext:
    """Context manager for streaming sweep data with live plotting.

    Accumulates data points via append() calls, streams to live plot writers
    immediately, and writes the complete sweep to file writers on context exit.

    Supports 1D sweeps (scalar x, scalar y) and 2D sweeps ([x0, x1], y).

    Example:
        >>> # 1D sweep
        >>> with session.sweep("signal", "Time", "Amplitude") as s:
        ...     for x, y in data_stream:
        ...         s.append([x], [y])  # Live plot updates

        >>> # 2D sweep
        >>> with session.sweep("heatmap", ["V1", "V2"], "Current") as s:
        ...     for v1, v2 in voltage_pairs:
        ...         s.append([v1, v2], [current])  # Live heatmap updates
        >>> # Sweep written to files here
    """

    def __init__(
        self,
        session: LoggerSession,
        name: str,
        x_label: str | list[str],
        y_label: str,
        metadata: dict[str, Any],
        routine_name: str | None = None,
    ):
        """Initialize sweep context.

        Args:
            session: Parent LoggerSession
            name: Sweep name
            x_label: X-axis label (string for 1D, list for 2D)
            y_label: Y-axis label
            metadata: User metadata dict
            routine_name: Optional routine name
        """
        self.session = session
        self.name = name
        self.x_label = x_label
        self.y_label = y_label
        # Defensive copy to avoid caller mutation affecting persisted record
        self.metadata = dict(metadata) if metadata else {}
        self.routine_name = routine_name

        self._x_data: list[float] | list[list[float]] = []
        self._y_data: list[float] = []
        self._dim: int | None = None  # Detect on first append
        self._active = True

    def append(
        self, x_data: list[float] | np.ndarray, y_data: list[float] | np.ndarray
    ) -> None:
        """Append data points and stream to live plots.

        1D: append([x], [y]) or append([x1, x2, ...], [y1, y2, ...])
        2D: append([x, y], [value])
        """
        if not self._active:
            raise ValueError(f"Sweep '{self.name}' is no longer active")

        x_arr = np.asarray(x_data, dtype=float)
        y_arr = np.asarray(y_data, dtype=float)

        if x_arr.size == 0 or y_arr.size == 0:
            raise ValueError("x_data and y_data cannot be empty")

        if x_arr.ndim != 1 or y_arr.ndim != 1:
            raise ValueError("x_data and y_data must be 1D arrays")

        if not (np.isfinite(x_arr).all() and np.isfinite(y_arr).all()):
            raise ValueError("x_data and y_data cannot contain NaN or inf")

        # Detect dimension on first append
        if self._dim is None:
            expecting_2d = isinstance(self.x_label, list) and len(self.x_label) == 2

            if len(y_arr) == 1:
                # Single y value: determine dim from x length
                self._dim = len(x_arr)
                if expecting_2d and self._dim != 2:
                    raise ValueError(
                        f"x_label has 2 elements but x_data has {self._dim}"
                    )
            elif len(x_arr) == len(y_arr):
                # Matching lengths: 1D batch mode
                if expecting_2d:
                    raise ValueError("For 2D sweeps, y_data must be length 1")
                self._dim = 1
            else:
                raise ValueError(
                    f"Ambiguous: x has {len(x_arr)} values, y has {len(y_arr)}. "
                    f"Use append([x], [y]) for 1D or append([x, y], [value]) for 2D."
                )

            if self._dim not in (1, 2):
                raise ValueError(f"Only 1D and 2D supported, got {self._dim}D")

            self.metadata["_dim"] = self._dim

        # Validate and accumulate
        if self._dim == 1:
            if len(x_arr) != len(y_arr):
                raise ValueError(f"Length mismatch: {len(x_arr)} vs {len(y_arr)}")
            self._x_data.extend(x_arr.tolist())
            self._y_data.extend(y_arr.tolist())
        else:  # 2D
            if len(x_arr) != 2 or len(y_arr) != 1:
                raise ValueError("2D requires len(x)==2, len(y)==1")
            self._x_data.append(x_arr.tolist())
            self._y_data.append(float(y_arr[0]))

        self.session._stream_live_sweep_chunk(
            self.name, x_arr, y_arr, self.x_label, self.y_label, self._dim
        )

    def end(self) -> None:
        """Write accumulated sweep to file writers."""
        if not self._active:
            return

        self._active = False
        self.session._active_sweeps.pop(self.name, None)

        if not self._x_data:
            logger.debug("Sweep '%s' empty, skipping write", self.name)
            return

        # Filter internal metadata (keys starting with _)
        user_metadata = {
            k: v for k, v in self.metadata.items() if not k.startswith("_")
        }

        sweep_data = SweepData(
            name=self.name,
            x_data=np.array(self._x_data, dtype=float),
            y_data=np.array(self._y_data, dtype=float),
            x_label=self.x_label,
            y_label=self.y_label,
            metadata=user_metadata,
            timestamp=time.time(),
            session_id=self.session.session_id,
            routine_name=self.routine_name,
        )

        self.session._write_completed_sweep(sweep_data)

    def cancel(self) -> None:
        """Cancel without persisting. Live updates remain visible."""
        self._active = False
        self.session._active_sweeps.pop(self.name, None)
        self._x_data.clear()
        self._y_data.clear()

    def __enter__(self) -> SweepContext:
        """Enter context manager."""
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: Any,
    ) -> None:
        """Exit context manager: end() on success, cancel() on exception."""
        if not self._active:
            return
        if exc_type is None:
            self.end()
        else:
            self.cancel()


class LoggerSession:
    """Session for the logger."""

    def __init__(
        self,
        metadata: SessionMetadata,
        writer_pool: dict[str, AbstractDataWriter],
        writer_refs: list[str],
        base_dir: str | Path,
        buffer_size: int = 1000,
        auto_flush_interval: float | None = 30.0,
    ):
        if not writer_pool or not writer_refs:
            raise LoggerSessionError("Writer pool and references are required")

        base_path = Path(base_dir)
        if not base_path.exists():
            raise LoggerSessionError(f"Base directory does not exist: {base_dir}")

        self.metadata = metadata
        self._writer_pool = writer_pool
        self._writer_refs = writer_refs
        self._base_dir = base_path
        self._buffer_size = buffer_size
        self._auto_flush_interval = auto_flush_interval

        self._active = False
        self._buffer: list[MeasurementData | SweepData] = []
        self._last_flush_time = time.monotonic()
        self._buffer_size_warning_threshold = buffer_size * 10
        self._buffer_size_warned = False
        self._live_writers: list[BokehLiveWriter] = []
        self._active_sweeps: dict[str, SweepContext] = {}

    @property
    def session_id(self) -> str:
        return self.metadata.session_id

    def _check_buffer_size_warning(self) -> None:
        """Check if buffer has grown too large and log warning once."""
        if (
            not self._buffer_size_warned
            and len(self._buffer) >= self._buffer_size_warning_threshold
        ):
            logger.warning(
                "Buffer size critical for session %s: %d items (threshold: %d). "
                "Consider flushing more frequently or increasing buffer size.",
                self.session_id,
                len(self._buffer),
                self._buffer_size_warning_threshold,
            )
            self._buffer_size_warned = True

    def initialize(self) -> None:
        """Initialize the session.

        Raises:
            LoggerSessionError: If session is already initialized
        """
        if self._active:
            raise LoggerSessionError("Session is already initialized")

        initialized_refs: list[str] = []
        try:
            for writer_ref in self._writer_refs:
                writer = self._writer_pool[writer_ref]
                writer.initialize_session(self.metadata)
                initialized_refs.append(writer_ref)

            # Cache live writers for streaming sweeps
            self._live_writers = [
                w
                for ref in self._writer_refs
                if isinstance(w := self._writer_pool[ref], BokehLiveWriter)
            ]

            self._active = True
            self._last_flush_time = time.monotonic()

        except Exception as e:
            # Cleanup any writers that initialized successfully
            for ref in initialized_refs:
                try:
                    self._writer_pool[ref].finalize_session(self.metadata)
                except Exception:
                    pass
            self._live_writers = []
            self._active = False
            raise LoggerSessionError(f"Failed to initialize session: {str(e)}") from e

    def finalize(self) -> None:
        """Finalize the session."""
        if not self._active:
            raise LoggerSessionError("Session is not initialized")

        try:
            if self._active_sweeps:
                logger.warning(
                    "Auto-completing %d active sweeps: %s",
                    len(self._active_sweeps),
                    list(self._active_sweeps.keys()),
                )
                for sweep in list(self._active_sweeps.values()):
                    try:
                        sweep.end()
                    except Exception as e:
                        logger.error(
                            "Failed to auto-complete sweep '%s': %s", sweep.name, e
                        )

            self.flush()

            for ref in self._writer_refs:
                self._writer_pool[ref].finalize_session(self.metadata)

            self._active = False

        except Exception as e:
            self._active = False
            raise LoggerSessionError(f"Failed to finalize session: {str(e)}") from e

    def close(self) -> None:
        """Close the session (alias for finalize())."""
        self.finalize()

    def log_measurement(
        self,
        name: str,
        data: dict[str, Any],
        metadata: dict[str, Any] | None = None,
        routine_name: str | None = None,
    ) -> None:
        """Log measurement data to a session."""
        if not self._active:
            raise LoggerSessionError("Session is not initialized")

        if not name or not name.strip():
            raise LoggerSessionError("Measurement name cannot be empty")

        if not data:
            raise LoggerSessionError("Measurement data cannot be empty")

        measurement = MeasurementData(
            name=name,
            data=data,
            metadata=metadata or {},
            timestamp=time.time(),
            session_id=self.session_id,
            routine_name=routine_name,
        )

        self._buffer.append(measurement)
        self._check_buffer_size_warning()

        if self._should_flush():
            self.flush()

    def log_analysis(
        self,
        name: str,
        data: dict[str, Any],
        metadata: dict[str, Any] | None = None,
        routine_name: str | None = None,
    ) -> None:
        """Log analysis data to a session."""
        meta = metadata.copy() if metadata else {}
        meta["data_type"] = "analysis"
        self.log_measurement(name, data, meta, routine_name)

    def log_sweep(
        self,
        name: str,
        x_data: list[float] | list[list[float]] | np.ndarray,
        y_data: list[float] | np.ndarray,
        x_label: str,
        y_label: str,
        metadata: dict[str, Any] | None = None,
    ) -> None:
        """Log sweep data to a session."""
        if not self._active:
            raise LoggerSessionError("Session is not initialized")

        if not name or not name.strip():
            raise LoggerSessionError("Sweep name cannot be empty")

        x_array = np.asarray(x_data)
        y_array = np.asarray(y_data)

        if x_array.size == 0 or y_array.size == 0:
            raise LoggerSessionError("Sweep data cannot be empty")

        sweep = SweepData(
            name=name,
            x_data=x_array,
            y_data=y_array,
            x_label=x_label,
            y_label=y_label,
            metadata=metadata or {},
            timestamp=time.time(),
            session_id=self.session_id,
        )

        self._buffer.append(sweep)
        self._check_buffer_size_warning()

        if self._should_flush():
            self.flush()

    def log_parameters(self, parameters: dict[str, Any]) -> None:
        """Log parameters to a session."""
        if not self._active:
            raise LoggerSessionError("Session is not initialized")

        if not parameters:
            raise LoggerSessionError("Parameters cannot be empty")

        if self.metadata.parameters is None:
            self.metadata.parameters = {}
        self.metadata.parameters.update(parameters)

    def sweep(
        self,
        name: str,
        x_label: str | list[str],
        y_label: str,
        metadata: dict[str, Any] | None = None,
    ) -> SweepContext:
        """Create streaming sweep context for live plotting.

        Accumulates data via append(), writes complete sweep to files on exit.
        Supports 1D and 2D data. Exception in context = no file write.
        One active sweep per name; HDF5 auto-uniquifies sequential same-name sweeps.

        Example:
            # 1D sweep
            with session.sweep("signal", "Time (s)", "Amplitude") as s:
                for x, y in data_stream:
                    s.append([x], [y])

            # 2D sweep
            with session.sweep("heatmap", ["V1 (V)", "V2 (V)"], "Current (A)") as s:
                for v1, v2 in voltage_pairs:
                    s.append([v1, v2], [current])
        """
        if not self._active:
            raise LoggerSessionError("Session is not initialized")

        if name in self._active_sweeps:
            raise LoggerSessionError(
                f"Sweep '{name}' is already active. Call end() or use a different name."
            )

        context = SweepContext(
            session=self,
            name=name,
            x_label=x_label,
            y_label=y_label,
            metadata=metadata or {},
            routine_name=self.metadata.routine_name,
        )

        self._active_sweeps[name] = context
        return context

    def _should_flush(self) -> bool:
        """Check if buffer should be flushed."""
        return (
            len(self._buffer) >= self._buffer_size
            or self._auto_flush_interval is not None
            and time.monotonic() - self._last_flush_time >= self._auto_flush_interval
        )

    def _stream_live_sweep_chunk(
        self,
        name: str,
        x: np.ndarray,
        y: np.ndarray,
        x_label: str | list[str],
        y_label: str,
        dim: int,
    ) -> None:
        """Stream data chunk to live plot writers only."""
        if not self._live_writers:
            return

        # Reshape x_data for consistent SweepData format
        x_data: np.ndarray
        if dim == 1:
            x_data = x.reshape(-1)
        else:  # dim == 2
            x_data = x.reshape(1, -1)

        temp_sweep = SweepData(
            name=name,
            x_data=x_data,
            y_data=y,
            x_label=x_label,
            y_label=y_label,
            metadata={"_dim": dim},
            timestamp=time.time(),
            session_id=self.session_id,
        )

        for writer in self._live_writers:
            try:
                writer.write_sweep(temp_sweep)
            except Exception as e:
                logger.warning("Live plot stream failed: %s", e)

    def _write_completed_sweep(self, sweep_data: SweepData) -> None:
        """Write complete sweep to file writers only (skip live/bokeh)."""
        for ref in self._writer_refs:
            if ref == "bokeh":
                continue
            try:
                self._writer_pool[ref].write_sweep(sweep_data)
                self._writer_pool[ref].flush()
            except Exception as e:
                logger.error("Failed to write sweep to %s: %s", ref, e)
                raise LoggerSessionError(f"Sweep write failed: {ref}: {e}") from e

    def flush(self) -> None:
        """Flush buffered data to all writers.

        Raises:
            LoggerSessionError: If any writer fails to write or flush data
        """
        if not self._buffer:
            return

        errors = []

        for item in self._buffer:
            for ref in self._writer_refs:
                writer = self._writer_pool[ref]
                try:
                    if isinstance(item, MeasurementData):
                        writer.write_measurement(item)
                    elif isinstance(item, SweepData):
                        writer.write_sweep(item)
                    else:
                        raise LoggerSessionError(f"Invalid item type: {type(item)}")
                except Exception as e:  # noqa: BLE001
                    logger.error("Write failed to %s: %s", ref, e)
                    errors.append(f"{ref}: {e}")

        for ref in self._writer_refs:
            try:
                self._writer_pool[ref].flush()
            except Exception as e:  # noqa: BLE001
                logger.error("Flush failed for %s: %s", ref, e)
                errors.append(f"{ref}: {e}")

        if errors:
            raise LoggerSessionError(f"Flush failed: {errors[0]}")

        self._buffer.clear()
        self._last_flush_time = time.monotonic()
        self._buffer_size_warned = False

    def __enter__(self) -> LoggerSession:
        """Enter the session context."""
        if not self._active:
            self.initialize()
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: Any,
    ) -> None:
        """Exit the session context."""
        try:
            if self._active:
                self.finalize()
        except Exception as e:
            logger.error("Failed to finalize session: %s", str(e))
            raise LoggerSessionError(f"Failed to finalize session: {str(e)}") from e

    def __repr__(self) -> str:
        return (
            f"LoggerSession(session_id={self.session_id}, "
            f"routine_name={self.metadata.routine_name}), active={self._active}, "
            f"buffer={len(self._buffer)}/{self._buffer_size}, "
            f"active_sweeps={len(self._active_sweeps)})"
        )



import importlib

from stanza.base.mixins import InstrumentChannelMixin
from stanza.base.protocols import (
    BreakoutBoxInstrument,
    ControlInstrument,
    MeasurementInstrument,
)
from stanza.models import InstrumentType


def load_driver_class(driver_name: str) -> type:
    module = importlib.import_module(f"stanza.drivers.{driver_name.lower()}")

    for attr_name in dir(module):
        attr = getattr(module, attr_name)
        if (
            isinstance(attr, type)
            and issubclass(attr, InstrumentChannelMixin)
            and attr.__module__ == module.__name__
        ):
            return attr

    raise ImportError(f"No instrument class found in stanza.drivers.{driver_name}")


def validate_driver_protocols(
    driver_class: type, instrument_type: InstrumentType
) -> None:
    dummy: object = object.__new__(driver_class)

    match instrument_type:
        case InstrumentType.CONTROL:
            if not isinstance(dummy, ControlInstrument):
                raise TypeError(
                    f"{driver_class.__name__} must implement ControlInstrument"
                )
        case InstrumentType.MEASUREMENT:
            if not isinstance(dummy, MeasurementInstrument):
                raise TypeError(
                    f"{driver_class.__name__} must implement MeasurementInstrument"
                )
        case InstrumentType.GENERAL:
            if not isinstance(dummy, ControlInstrument):
                raise TypeError(
                    f"{driver_class.__name__} must implement ControlInstrument"
                )
            if not isinstance(dummy, MeasurementInstrument):
                raise TypeError(
                    f"{driver_class.__name__} must implement MeasurementInstrument"
                )
        case InstrumentType.BREAKOUT_BOX:
            if not isinstance(dummy, BreakoutBoxInstrument):
                raise TypeError(
                    f"{driver_class.__name__} must implement BreakoutBoxInstrument"
                )



"""Data writer for live plotting with Bokeh."""

from __future__ import annotations

from typing import Any

from stanza.logger.datatypes import MeasurementData, SessionMetadata, SweepData
from stanza.logger.writers.base import AbstractDataWriter


class BokehLiveWriter(AbstractDataWriter):
    """Stream sweep data to Bokeh plots."""

    def __init__(self, backend: Any, max_points: int = 1000) -> None:
        """
        Args:
            backend: ServerBackend or InlineBackend instance
            max_points: Maximum points per plot (older data rolls off for 1D)
        """
        self.backend = backend
        self.max_points = max_points
        self._plots: dict[str, dict[str, Any]] = {}  # Store plot metadata
        self._initialized: bool = False

    def initialize_session(self, metadata: SessionMetadata) -> None:
        """Start of new session."""
        self._initialized = True

    def write_measurement(self, data: MeasurementData) -> None:
        """Write single measurement (not used for plotting)."""
        pass

    def write_sweep(self, data: SweepData) -> None:
        """Stream sweep data to plot."""
        dim = data.metadata.get("_dim") or (
            1 if data.x_data.ndim == 1 else data.x_data.shape[1]
        )

        if dim not in (1, 2):
            raise ValueError(f"Only 1D and 2D supported, got {dim}D")

        if data.name not in self._plots:
            (self._create_1d_plot if dim == 1 else self._create_2d_plot)(data)

        (self._stream_1d if dim == 1 else self._stream_2d)(data)

    def _create_1d_plot(self, data: SweepData) -> None:
        """Create 1D line plot."""
        self.backend.create_figure(
            name=data.name,
            x_label=data.x_label if isinstance(data.x_label, str) else "X",
            y_label=data.y_label or "Y",
            plot_type="line",
        )
        self._plots[data.name] = {"dim": 1}

    def _create_2d_plot(self, data: SweepData) -> None:
        """Create 2D heatmap using rect glyph."""
        labels = data.x_label if isinstance(data.x_label, list) else ["X", "Y"]
        x_label = labels[0] if len(labels) > 0 else "X"
        y_label = labels[1] if len(labels) > 1 else "Y"

        self.backend.create_figure(
            name=data.name,
            x_label=x_label,
            y_label=y_label,
            z_label=data.y_label or "Value",
            plot_type="heatmap",
            cell_size=data.metadata.get("cell_size"),
        )
        self._plots[data.name] = {"dim": 2}

    def _stream_1d(self, data: SweepData) -> None:
        """Stream 1D data to line plot."""
        self.backend.stream_data(
            name=data.name,
            new_data={"x": list(data.x_data), "y": list(data.y_data)},
            rollover=self.max_points,
        )

    def _stream_2d(self, data: SweepData) -> None:
        """Stream 2D data to heatmap."""
        self.backend.stream_data(
            name=data.name,
            new_data={
                "x": data.x_data[:, 0].tolist(),
                "y": data.x_data[:, 1].tolist(),
                "value": data.y_data.tolist(),
            },
            rollover=None,
        )

    def flush(self) -> None:
        """Flush any pending updates."""
        if hasattr(self.backend, "push_updates"):
            self.backend.push_updates()

    def finalize_session(self, metadata: SessionMetadata | None = None) -> None:
        """End of session."""
        self.flush()


__all__ = ["BokehLiveWriter"]



"""Data writers for logging."""

from stanza.logger.writers.base import AbstractDataWriter
from stanza.logger.writers.bokeh_writer import BokehLiveWriter
from stanza.logger.writers.hdf5_writer import HDF5Writer
from stanza.logger.writers.jsonl_writer import JSONLWriter

__all__ = ["AbstractDataWriter", "BokehLiveWriter", "HDF5Writer", "JSONLWriter"]



"""Server-based plotting backend for live visualization in browser.

Runs a Bokeh server in a background thread, streams data updates via WebSocket.
Works from any environment: scripts, notebooks, or interactive sessions.
"""

from __future__ import annotations

import asyncio
import logging
import threading
import time
from typing import Any

from bokeh.application import Application
from bokeh.application.handlers.function import FunctionHandler
from bokeh.server.server import Server

from stanza.plotter.backends.utils import (
    PlotSpec,
    PlotState,
    make_heatmap_plot,
    make_line_plot,
    prepare_heatmap_data,
)

logger = logging.getLogger(__name__)


class ServerBackend:
    """Live plotting in browser via Bokeh server.

    Runs embedded server in background thread. Server dies when parent process exits.
    """

    def __init__(self, port: int = 5006, daemon: bool = True) -> None:
        self.port = port
        self.daemon = daemon
        self._server: Server | None = None
        self._doc: Any = None
        self._running = False

        # Plot configurations registered before browser connects
        self._plot_specs: dict[str, PlotSpec] = {}
        # Complete plot states after browser connects and plots are created
        self._plots: dict[str, PlotState] = {}
        # Data buffered before browser connects
        self._buffer: dict[str, dict[str, list[Any]]] = {}

    def start(self, block: bool = False) -> None:
        """Start Bokeh server.

        Args:
            block: If True, blocks until server is stopped (for persistent mode).
                   If False, runs in background thread (for embedded mode).
        """
        if self._running:
            return

        logger.info(f"Instantiating new Bokeh server on port {self.port}")

        def make_document(doc: Any) -> None:
            """Initialize document when browser connects."""
            self._doc = doc

            for name in self._plot_specs:
                self._create_plot(name)

        def run_server() -> None:
            """Server thread: create event loop and start server."""
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)

            app = Application(FunctionHandler(make_document))
            self._server = Server(
                {"/": app},
                port=self.port,
                allow_websocket_origin=[f"localhost:{self.port}"],
            )

            self._server.start()
            self._server.io_loop.start()

        if block or not self.daemon:
            # Run in main thread (blocks)
            self._running = True
            run_server()
        else:
            # Run in background daemon thread
            thread = threading.Thread(target=run_server, daemon=True)
            thread.start()
            self._running = True
            time.sleep(1.0)  # Give server time to start

    def stop(self) -> None:
        """Stop the Bokeh server and remove from active servers registry."""
        if self._server and self._running:
            self._server.unlisten()
            self._server.stop()
            self._server.io_loop.stop()
            self._running = False

            from stanza.plotter import _active_servers

            if self.port in _active_servers and _active_servers[self.port] is self:
                del _active_servers[self.port]

    def reset(self) -> None:
        """Reset the document, clearing all plots and data sources.

        This allows reusing the same server instance across multiple sessions.
        """
        if not self._running:
            return

        self._plots.clear()
        self._plot_specs.clear()
        self._buffer.clear()

        # If browser is connected, clear the document too
        if self._doc is not None:

            def do_clear() -> None:
                if self._doc is not None:
                    self._doc.clear()

            self._doc.add_next_tick_callback(do_clear)

    def create_figure(
        self,
        name: str,
        x_label: str,
        y_label: str,
        plot_type: str = "line",
        z_label: str | None = None,
        cell_size: tuple[float, float] | None = None,
    ) -> None:
        """Register a new plot. Created when browser connects."""
        if not self._running:
            raise RuntimeError("Server not started")

        if name in self._plot_specs:
            return

        # Store partial spec; will be replaced with full spec when plot is created
        self._plot_specs[name] = PlotSpec(
            name=name,
            plot_type=plot_type,
            x_label=x_label,
            y_label=y_label,
            z_label=z_label,
            cell_size=cell_size,
        )

        # If browser already connected, create plot immediately
        if self._doc is not None and name not in self._plots:
            self._doc.add_next_tick_callback(lambda: self._create_plot(name))

    def _create_plot(self, name: str) -> None:
        """Create plot based on spec."""
        spec = self._plot_specs[name]

        if spec.plot_type == "line":
            self._create_line_plot(name)
        elif spec.plot_type == "heatmap":
            self._create_heatmap_plot(name)
        else:
            raise ValueError(f"Unknown plot type: {spec.plot_type}")

    def _create_line_plot(self, name: str) -> None:
        """Create 1D line plot."""
        spec = self._plot_specs[name]
        data = self._buffer.pop(name, {"x": [], "y": []})
        plot_state = make_line_plot(name, spec.x_label, spec.y_label)
        if data["x"] or data["y"]:
            plot_state.source.data = data
        self._plots[name] = plot_state
        if self._doc:
            self._doc.add_root(plot_state.figure)

    def _create_heatmap_plot(self, name: str) -> None:
        """Create 2D heatmap with rect glyph and linear color mapping."""
        spec = self._plot_specs[name]
        data = self._buffer.pop(
            name, {"x": [], "y": [], "value": [], "width": [], "height": []}
        )
        plot_state = make_heatmap_plot(
            name, spec.x_label, spec.y_label, spec.z_label, spec.cell_size
        )
        if any(data.get(k) for k in ("x", "y", "value", "width", "height")):
            plot_state.source.data = data
        self._plots[name] = plot_state
        if self._doc:
            self._doc.add_root(plot_state.figure)

    def stream_data(
        self, name: str, new_data: dict[str, Any], rollover: int | None = None
    ) -> None:
        """Add data to plot. Buffers if browser not yet connected."""
        # Try to get created plot first
        plot = self._plots.get(name)

        if plot is not None:
            # Plot exists, stream to it
            if plot.spec.plot_type == "heatmap" and "value" in new_data:
                new_data = prepare_heatmap_data(new_data, plot.source.data, plot.spec)

            # Stream to existing plot (thread-safe via callback)
            if self._doc:

                def do_stream() -> None:
                    plot.source.stream(new_data, rollover=rollover)

                    # Update color mapper for heatmaps
                    if plot.spec.plot_type == "heatmap" and plot.spec.mapper:
                        plot.spec.mapper.low = plot.spec.value_min
                        plot.spec.mapper.high = plot.spec.value_max

                self._doc.add_next_tick_callback(do_stream)
        else:
            # Plot not yet created, buffer the data
            spec = self._plot_specs.get(name)
            if spec is None:
                return

            if spec.plot_type == "heatmap" and "value" in new_data:
                # Initialize heatmap spec fields if buffering data before plot creation
                if spec.dx is None and spec.cell_size:
                    spec.dx = spec.cell_size[0]
                if spec.dy is None and spec.cell_size:
                    spec.dy = spec.cell_size[1]
                new_data = prepare_heatmap_data(new_data, {}, spec)

            if name not in self._buffer:
                self._buffer[name] = {k: [] for k in new_data.keys()}
            for key, values in new_data.items():
                self._buffer[name].setdefault(key, []).extend(values)



from dataclasses import dataclass
from typing import Any

import numpy as np


@dataclass
class MeasurementData:
    """Represents a single measurement point."""

    name: str
    data: dict[str, Any]
    metadata: dict[str, Any]
    timestamp: float
    session_id: str
    routine_name: str | None = None

    def to_dict(self) -> dict[str, Any]:
        """Convert to dictionary for serialization."""
        return {
            "name": self.name,
            "data": self.data,
            "metadata": self.metadata,
            "timestamp": self.timestamp,
            "session_id": self.session_id,
            "routine_name": self.routine_name,
        }


@dataclass
class SweepData:
    """Represents a sweep of measurement data.

    Supports both 1D and 2D sweeps:
    - 1D: x_data.shape = (N,), y_data.shape = (N,)
    - 2D: x_data.shape = (N, 2), y_data.shape = (N,)
    """

    name: str
    x_data: np.ndarray
    y_data: np.ndarray
    x_label: str | list[str]
    y_label: str
    metadata: dict[str, Any]
    timestamp: float
    session_id: str
    routine_name: str | None = None

    def __post_init__(self) -> None:
        """Validate sweep data dimensions."""
        if self.y_data.ndim != 1:
            raise ValueError(f"y_data must be 1D, got shape {self.y_data.shape}")

        if len(self.x_data) != len(self.y_data):
            raise ValueError(
                f"Length mismatch: x_data={len(self.x_data)}, y_data={len(self.y_data)}"
            )

        if self.x_data.ndim == 2 and isinstance(self.x_label, list):
            if len(self.x_label) != self.x_data.shape[1]:
                raise ValueError(
                    f"x_label length {len(self.x_label)} != x_data dims {self.x_data.shape[1]}"
                )

    def to_dict(self) -> dict[str, Any]:
        """Convert to dictionary for serialization."""
        return {
            "name": self.name,
            "x_data": self.x_data.tolist(),
            "y_data": self.y_data.tolist(),
            "x_label": self.x_label,
            "y_label": self.y_label,
            "metadata": self.metadata,
            "timestamp": self.timestamp,
            "session_id": self.session_id,
            "routine_name": self.routine_name,
        }


@dataclass
class SessionMetadata:
    """Session-level metadata."""

    session_id: str
    start_time: float
    user: str
    routine_name: str | None = None
    device_config: dict[str, Any] | None = None
    parameters: dict[str, Any] | None = None
    end_time: float | None = None
    git_commit: str | None = None

    @property
    def duration(self) -> float | None:
        """Session duration in seconds."""
        return None if self.end_time is None else self.end_time - self.start_time

    def to_dict(self) -> dict[str, Any]:
        """Convert to dictionary for serialization."""
        return {
            "session_id": self.session_id,
            "routine_name": self.routine_name,
            "start_time": self.start_time,
            "user": self.user,
            "device_config": self.device_config,
            "parameters": self.parameters,
            "end_time": self.end_time,
            "git_commit": self.git_commit,
            "duration": self.duration,
        }



import importlib.resources
from importlib.resources import as_file

import yaml

from stanza.models import DeviceConfig
from stanza.utils import device_from_config, device_from_yaml


def test_device_from_yaml():
    with as_file(
        importlib.resources.files("tests.test_qdac2_pyvisa_sim").joinpath(
            "qdac2_pyvisa_sim.yaml"
        )
    ) as sim_file:
        device = device_from_yaml(
            "devices/device.sample.yaml",
            is_stanza_config=True,
            is_simulation=True,
            sim_file=str(sim_file),
        )

        assert device.name == "Sample Device"
        assert device.control_instrument is not None
        assert device.measurement_instrument is not None
        assert len(device.gates) == 3
        assert len(device.contacts) == 2


def test_device_from_config_with_qswitch(device_yaml_with_qswitch):
    """Test that qswitch breakout box can be loaded with device_from_config."""
    config = DeviceConfig.model_validate(yaml.safe_load(device_yaml_with_qswitch))

    with as_file(
        importlib.resources.files("tests.test_qdac2_pyvisa_sim").joinpath(
            "qdac2_pyvisa_sim.yaml"
        )
    ) as sim_file:
        device = device_from_config(config, is_simulation=True, sim_file=str(sim_file))

        assert device.breakout_box_instrument is not None
        assert {"G1", "C1"} <= device.breakout_box_instrument.channels.keys()
